import datetime
from time import sleep, time
from unittest.case import TestCase
from uuid import uuid4

from eventsourcing.domain.model.entity import VersionedEntity, TimestampedEntity
from eventsourcing.domain.model.events import DomainEvent
from eventsourcing.utils.times import decimaltimestamp
from eventsourcing.utils.topic import get_topic
from eventsourcing.infrastructure.sequenceditem import SequencedItem
from eventsourcing.infrastructure.sequenceditemmapper import SequencedItemMapper


class Event1(VersionedEntity.Event):
    pass


class Event2(TimestampedEntity.Event):
    pass


class Event3(DomainEvent):
    pass


class ValueObject1(object):
    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return self.__dict__ != other.__dict__


class TestSequencedItemMapper(TestCase):
    def test_with_versioned_entity_event(self):
        # Setup the mapper, and create an event.
        mapper = SequencedItemMapper(
            sequenced_item_class=SequencedItem,
            sequence_id_attr_name='originator_id',
            position_attr_name='originator_version'
        )
        entity_id1 = uuid4()
        event1 = Event1(originator_id=entity_id1, originator_version=101)

        # Check to_sequenced_item() method results in a sequenced item.
        sequenced_item = mapper.item_from_event(event1)
        self.assertIsInstance(sequenced_item, SequencedItem)
        self.assertEqual(sequenced_item.position, 101)
        self.assertEqual(sequenced_item.sequence_id, entity_id1)
        self.assertEqual(sequenced_item.topic, get_topic(Event1))
        self.assertTrue(sequenced_item.state)

        # Use the returned values to create a new sequenced item.
        sequenced_item_copy = SequencedItem(
            sequence_id=sequenced_item.sequence_id,
            position=sequenced_item.position,
            topic=sequenced_item.topic,
            state=sequenced_item.state,
        )

        # Check from_sequenced_item() returns an event.
        domain_event = mapper.event_from_item(sequenced_item_copy)
        self.assertIsInstance(domain_event, Event1)
        self.assertEqual(domain_event.originator_id, event1.originator_id)
        self.assertEqual(domain_event.originator_version, event1.originator_version)

    def test_with_timestamped_entity_event(self):
        # Setup the mapper, and create an event.
        mapper = SequencedItemMapper(
            sequenced_item_class=SequencedItem,
            sequence_id_attr_name='originator_id',
            position_attr_name='timestamp'
        )
        before = decimaltimestamp()
        sleep(0.000001)  # Avoid test failing due to timestamp having limited precision.
        event2 = Event2(originator_id='entity2')
        sleep(0.000001)  # Avoid test failing due to timestamp having limited precision.
        after = decimaltimestamp()

        # Check to_sequenced_item() method results in a sequenced item.
        sequenced_item = mapper.item_from_event(event2)
        self.assertIsInstance(sequenced_item, SequencedItem)
        self.assertGreater(sequenced_item.position, before)
        self.assertLess(sequenced_item.position, after)
        self.assertEqual(sequenced_item.sequence_id, 'entity2')
        self.assertEqual(sequenced_item.topic, get_topic(Event2))
        self.assertTrue(sequenced_item.state)

        # Use the returned values to create a new sequenced item.
        sequenced_item_copy = SequencedItem(
            sequence_id=sequenced_item.sequence_id,
            position=sequenced_item.position,
            topic=sequenced_item.topic,
            state=sequenced_item.state,
        )

        # Check from_sequenced_item() returns an event.
        domain_event = mapper.event_from_item(sequenced_item_copy)
        self.assertIsInstance(domain_event, Event2)
        self.assertEqual(domain_event.originator_id, event2.originator_id)
        self.assertEqual(domain_event.timestamp, event2.timestamp)

    def test_with_different_types_of_event_attributes(self):
        # Setup the mapper, and create an event.
        mapper = SequencedItemMapper(
            sequenced_item_class=SequencedItem,
            sequence_id_attr_name='originator_id',
            position_attr_name='a'
        )

        # Check value objects can be compared ok.
        self.assertEqual(ValueObject1('value1'), ValueObject1('value1'))
        self.assertNotEqual(ValueObject1('value1'), ValueObject1('value2'))

        # Create an event with dates and datetimes.
        event3 = Event3(
            originator_id='entity3',
            originator_version=303,
            a=datetime.datetime(2017, 3, 22, 9, 12, 14),
            b=datetime.date(2017, 3, 22),
            c=uuid4(),
            # d=Decimal(1.1),
            e=ValueObject1('value1'),
        )

        # Check to_sequenced_item() method results in a sequenced item.
        sequenced_item = mapper.item_from_event(event3)

        # Use the returned values to create a new sequenced item.
        sequenced_item_copy = SequencedItem(
            sequence_id=sequenced_item.sequence_id,
            position=sequenced_item.position,
            topic=sequenced_item.topic,
            state=sequenced_item.state,
        )

        # Check from_sequenced_item() returns an event.
        domain_event = mapper.event_from_item(sequenced_item_copy)
        self.assertIsInstance(domain_event, Event3)
        self.assertEqual(domain_event.originator_id, event3.originator_id)
        self.assertEqual(domain_event.a, event3.a)
        self.assertEqual(domain_event.b, event3.b)
        self.assertEqual(domain_event.c, event3.c)
        # self.assertEqual(domain_event.d, event3.d)
        self.assertEqual(domain_event.e, event3.e)



from eventsourcing.infrastructure.factory import InfrastructureFactory
from eventsourcing.infrastructure.popo.manager import PopoRecordManager


class PopoInfrastructureFactory(InfrastructureFactory):
    record_manager_class = PopoRecordManager



from uuid import UUID

from eventsourcing.application.policies import PersistencePolicy
from eventsourcing.example.domainmodel import create_new_example
from eventsourcing.example.infrastructure import ExampleRepository
from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.infrastructure.sequenceditem import StoredEvent
from eventsourcing.infrastructure.sequenceditemmapper import SequencedItemMapper
from eventsourcing.infrastructure.sqlalchemy.records import StoredEventRecord, Base
from eventsourcing.infrastructure.sqlalchemy.manager import SQLAlchemyRecordManager
from eventsourcing.infrastructure.sqlalchemy.datastore import SQLAlchemyDatastore, SQLAlchemySettings
from eventsourcing.tests.datastore_tests.base import AbstractDatastoreTestCase


# This test replaces the default SequencedItem class with a StoredEvent class.
# How easy is it to customize the infrastructure to support that? We just need
# to define the new sequenced item class, define a suitable record class,
# and configure the other components. It's easy.


class ExampleApplicationWithAlternativeSequencedItemType(object):
    def __init__(self, session):
        self.event_store = EventStore(
            record_manager=SQLAlchemyRecordManager(
                session=session,
                record_class=StoredEventRecord,
                sequenced_item_class=StoredEvent,
            ),
            sequenced_item_mapper=SequencedItemMapper(
                sequenced_item_class=StoredEvent,
                sequence_id_attr_name='originator_id',
                position_attr_name='originator_version',
            )
        )
        self.repository = ExampleRepository(
            event_store=self.event_store,
        )
        self.persistence_policy = PersistencePolicy(self.event_store)

    def close(self):
        self.persistence_policy.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class TestExampleWithAlternativeSequencedItemType(AbstractDatastoreTestCase):
    def setUp(self):
        super(TestExampleWithAlternativeSequencedItemType, self).setUp()
        self.datastore.setup_connection()
        self.datastore.drop_tables()  # something isn't dropping tables...
        self.datastore.setup_tables()

    def tearDown(self):
        self.datastore.drop_tables()
        self.datastore.close_connection()
        super(TestExampleWithAlternativeSequencedItemType, self).setUp()

    def construct_datastore(self):
        return SQLAlchemyDatastore(
            base=Base,
            settings=SQLAlchemySettings(),
            tables=(StoredEventRecord,)
        )

    def _test(self):
        with ExampleApplicationWithAlternativeSequencedItemType(self.datastore.session) as app:
            # Create entity.
            entity1 = create_new_example(a='a', b='b')
            self.assertIsInstance(entity1.id, UUID)
            self.assertEqual(entity1.a, 'a')
            self.assertEqual(entity1.b, 'b')

            # Check there is a stored event.
            all_records = list(app.event_store.record_manager.get_notifications())
            self.assertEqual(1, len(all_records))
            stored_event = all_records[0]
            self.assertEqual(stored_event.originator_id, entity1.id)
            self.assertEqual(stored_event.originator_version, 0)

            # Read entity from repo.
            retrieved_obj = app.repository[entity1.id]
            self.assertEqual(retrieved_obj.id, entity1.id)



from django.db import models


class IntegerSequencedRecord(models.Model):

    id = models.BigAutoField(primary_key=True)

    # Sequence ID (e.g. an entity or aggregate ID).
    sequence_id = models.UUIDField()

    # Position (index) of item in sequence.
    position = models.BigIntegerField()

    # Topic of the item (e.g. path to domain event class).
    topic = models.TextField()

    # State of the item (serialized dict, possibly encrypted).
    state = models.TextField()

    class Meta:
        unique_together = (("sequence_id", "position"),)
        db_table = 'integer_sequenced_items'


class TimestampSequencedRecord(models.Model):

    id = models.BigAutoField(primary_key=True)

    # Sequence ID (e.g. an entity or aggregate ID).
    sequence_id = models.UUIDField()

    # Position (timestamp) of item in sequence.
    position = models.DecimalField(max_digits=24, decimal_places=6)

    # Topic of the item (e.g. path to domain event class).
    topic = models.TextField()

    # State of the item (serialized dict, possibly encrypted).
    state = models.TextField()

    class Meta:
        unique_together = (("sequence_id", "position"),)
        db_table = 'timestamp_sequenced_items'
        indexes = [
            models.Index(fields=['position'], name='position_idx'),
        ]


class SnapshotRecord(models.Model):

    uid = models.BigAutoField(primary_key=True)

    # Sequence ID (e.g. an entity or aggregate ID).
    sequence_id = models.UUIDField()

    # Position (index) of item in sequence.
    position = models.BigIntegerField()

    # Topic of the item (e.g. path to domain event class).
    topic = models.TextField()

    # State of the item (serialized dict, possibly encrypted).
    state = models.TextField()

    class Meta:
        unique_together = (("sequence_id", "position"),)
        db_table = 'snapshots'


class EntitySnapshotRecord(models.Model):

    uid = models.BigAutoField(primary_key=True)

    # Application name.
    application_name = models.CharField(max_length=32)

    # Originator ID (e.g. an entity or aggregate ID).
    originator_id = models.UUIDField()

    # Originator version of item in sequence.
    originator_version = models.BigIntegerField()

    # Topic of the item (e.g. path to domain event class).
    topic = models.TextField()

    # State of the item (serialized dict, possibly encrypted).
    state = models.TextField()

    class Meta:
        unique_together = (
            ("application_name", "originator_id", "originator_version"),
        )
        db_table = 'entity_snapshots'


class StoredEventRecord(models.Model):

    uid = models.BigAutoField(primary_key=True)

    # Application name.
    application_name = models.CharField(max_length=32)

    # Sequence ID (e.g. an entity or aggregate ID).
    originator_id = models.UUIDField()

    # Position (index) of item in sequence.
    originator_version = models.BigIntegerField()

    # Upstream application name.
    pipeline_id = models.IntegerField()

    # Notification ID.
    notification_id = models.BigIntegerField()

    # Topic of the item (e.g. path to domain event class).
    topic = models.TextField()

    # State of the item (serialized dict, possibly encrypted).
    state = models.TextField()

    # Causal dependencies.
    causal_dependencies = models.TextField()

    class Meta:
        unique_together = (
            ("application_name", "originator_id", "originator_version"),
            ("application_name", "pipeline_id", "notification_id"),
        )
        db_table = 'stored_events'


class NotificationTrackingRecord(models.Model):

    uid = models.BigAutoField(primary_key=True)

    # Application name.
    application_name = models.CharField(max_length=32)

    # Upstream application name.
    upstream_application_name = models.CharField(max_length=32)

    # Pipeline ID.
    pipeline_id = models.IntegerField()

    # Notification ID.
    notification_id = models.BigIntegerField()

    class Meta:
        unique_together = (
            (
                "application_name",
                "upstream_application_name",
                "pipeline_id",
                "notification_id"
            ),
        )
        db_table = 'notification_tracking'



from abc import ABC, abstractmethod
from copy import deepcopy

from eventsourcing.domain.model.snapshot import AbstractSnapshop, Snapshot
from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.infrastructure.sequenceditemmapper import reconstruct_object
from eventsourcing.utils.topic import get_topic, resolve_topic


class AbstractSnapshotStrategy(ABC):
    @abstractmethod
    def get_snapshot(self, entity_id, lt=None, lte=None):
        """
        Gets the last snapshot for entity, optionally until a particular version number.

        :rtype: Snapshot
        """

    @abstractmethod
    def take_snapshot(self, entity_id, entity, last_event_version):
        """
        Takes a snapshot of entity, using given ID, state and version number.

        :rtype: AbstractSnapshop
        """


class EventSourcedSnapshotStrategy(AbstractSnapshotStrategy):
    """Snapshot strategy that uses an event sourced snapshot.
    """

    def __init__(self, snapshot_store):
        assert isinstance(snapshot_store, EventStore)
        self.snapshot_store = snapshot_store

    def get_snapshot(self, entity_id, lt=None, lte=None):
        """
        Gets the last snapshot for entity, optionally until a particular version number.

        :rtype: Snapshot
        """
        snapshots = self.snapshot_store.get_domain_events(entity_id, lt=lt, lte=lte, limit=1, is_ascending=False)
        if len(snapshots) == 1:
            return snapshots[0]

    # Todo: Rename as create_snapshot?
    def take_snapshot(self, entity_id, entity, last_event_version):
        """
        Creates a Snapshot from the given state, and appends it
        to the snapshot store.

        :rtype: Snapshot
        """

        # Create the snapshot.
        snapshot = Snapshot(
            originator_id=entity_id,
            originator_version=last_event_version,
            topic=get_topic(entity.__class__),
            state=None if entity is None else deepcopy(entity.__dict__)
        )

        self.snapshot_store.store(snapshot)

        # Return the snapshot.
        return snapshot


def entity_from_snapshot(snapshot):
    """
    Reconstructs domain entity from given snapshot.
    """
    assert isinstance(snapshot, AbstractSnapshop), type(snapshot)
    if snapshot.state is not None:
        entity_class = resolve_topic(snapshot.topic)
        return reconstruct_object(entity_class, snapshot.state)



from eventsourcing.application.simple import ApplicationWithConcreteInfrastructure
from eventsourcing.infrastructure.sqlalchemy.factory import SQLAlchemyInfrastructureFactory
from eventsourcing.infrastructure.sqlalchemy.records import EntitySnapshotRecord, StoredEventRecord


class SQLAlchemyApplication(ApplicationWithConcreteInfrastructure):
    infrastructure_factory_class = SQLAlchemyInfrastructureFactory
    stored_event_record_class = StoredEventRecord
    snapshot_record_class = EntitySnapshotRecord
    is_constructed_with_session = True
    tracking_record_class = None

    def __init__(self, uri=None, session=None, tracking_record_class=None, **kwargs):
        self.uri = uri
        self.session = session
        self.tracking_record_class = tracking_record_class or type(self).tracking_record_class
        super(SQLAlchemyApplication, self).__init__(**kwargs)

    def construct_infrastructure(self, *args, **kwargs):
        super(SQLAlchemyApplication, self).construct_infrastructure(
            session=self.session, uri=self.uri,
            tracking_record_class=self.tracking_record_class,
            *args, **kwargs
        )
        if self.datastore and self.session is None:
            self.session = self.datastore.session



from eventsourcing.infrastructure.django.manager import DjangoRecordManager
from eventsourcing.infrastructure.factory import InfrastructureFactory


class DjangoInfrastructureFactory(InfrastructureFactory):
    """
    Infrastructure factory for Django.
    """
    record_manager_class = DjangoRecordManager
    tracking_record_class = None

    def __init__(self, tracking_record_class=None, *args, **kwargs):
        super(DjangoInfrastructureFactory, self).__init__(*args, **kwargs)
        self._tracking_record_class = tracking_record_class

    @property
    def integer_sequenced_record_class(self):
        from eventsourcing.infrastructure.django.models import IntegerSequencedRecord
        return IntegerSequencedRecord

    @property
    def timestamp_sequenced_record_class(self):
        from eventsourcing.infrastructure.django.models import TimestampSequencedRecord
        return TimestampSequencedRecord

    @property
    def snapshot_record_class(self):
        from eventsourcing.infrastructure.django.models import SnapshotRecord
        return SnapshotRecord

    def construct_integer_sequenced_record_manager(self, **kwargs):
        """
        Constructs Django record manager.

        :return: A Django record manager.
        :rtype: DjangoRecordManager
        """
        tracking_record_class = self._tracking_record_class or self.tracking_record_class
        return super(DjangoInfrastructureFactory, self).construct_integer_sequenced_record_manager(
            tracking_record_class=tracking_record_class,
            **kwargs
        )



from collections import namedtuple
from uuid import UUID

from sqlalchemy import DECIMAL
from sqlalchemy.sql.schema import Column, Index
from sqlalchemy.sql.sqltypes import BigInteger, Integer, String, Text
from sqlalchemy_utils.types.uuid import UUIDType

from eventsourcing.application.policies import PersistencePolicy
from eventsourcing.domain.model.events import DomainEvent
from eventsourcing.example.domainmodel import create_new_example
from eventsourcing.example.infrastructure import ExampleRepository
from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.infrastructure.sequenceditemmapper import SequencedItemMapper
from eventsourcing.infrastructure.sqlalchemy.datastore import SQLAlchemyDatastore, SQLAlchemySettings
from eventsourcing.infrastructure.sqlalchemy.manager import SQLAlchemyRecordManager
from eventsourcing.infrastructure.sqlalchemy.records import Base
from eventsourcing.tests.datastore_tests.base import AbstractDatastoreTestCase

# This module explores extending the sequenced item class with some more fields. How easy is it?
# Just needed to define the extended type, define a suitable record
# class, and extend the sequenced itemevent mapper to derive values for the
# extra attributes. It's easy.

# Define the sequenced item class.
ExtendedSequencedItem = namedtuple(
    'ExtendedSequencedItem',
    ['sequence_id', 'position', 'topic', 'state', 'timestamp', 'event_type']
)


# Extend the database table definition to support the extra fields.
class ExtendedIntegerSequencedRecord(Base):
    __tablename__ = 'extended_integer_sequenced_items'

    id = Column(BigInteger().with_variant(Integer, "sqlite"), primary_key=True)

    # Sequence ID (e.g. an entity or aggregate ID).
    sequence_id = Column(UUIDType(), nullable=False)

    # Position (index) of item in sequence.
    position = Column(BigInteger(), nullable=False)

    # Topic of the item (e.g. path to domain event class).
    topic = Column(String(255), nullable=False)

    # State of the item (serialized dict, possibly encrypted).
    state = Column(Text())

    # Timestamp of the event.
    timestamp = Column(DECIMAL(24, 6, 6), nullable=False)
    # timestamp = Column(DECIMAL(27, 9, 9), nullable=False)

    # Type of the event (class name).
    event_type = Column(String(255))

    __table_args__ = (
        Index('integer_sequenced_items_index', 'sequence_id', 'position', unique=True),
    )


# Extend the sequenced item mapper to derive the extra values.
class ExtendedSequencedItemMapper(SequencedItemMapper):
    def construct_item_args(self, domain_event):
        args = super(ExtendedSequencedItemMapper, self).construct_item_args(domain_event)
        event_type = domain_event.__class__.__qualname__
        return args + (event_type,)


# Define an application object.
class ExampleApplicationWithExtendedSequencedItemType(object):
    def __init__(self, session):
        self.event_store = EventStore(
            record_manager=SQLAlchemyRecordManager(
                session=session,
                record_class=ExtendedIntegerSequencedRecord,
                sequenced_item_class=ExtendedSequencedItem,
            ),
            sequenced_item_mapper=ExtendedSequencedItemMapper(
                sequenced_item_class=ExtendedSequencedItem,
                sequence_id_attr_name='originator_id',
                position_attr_name='originator_version',
                other_attr_names=('timestamp',),
            )
        )
        self.repository = ExampleRepository(
            event_store=self.event_store,
        )
        self.persistence_policy = PersistencePolicy(
            event_store=self.event_store,
            persist_event_type=DomainEvent,
        )

    def close(self):
        self.persistence_policy.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class TestExampleWithExtendedSequencedItemType(AbstractDatastoreTestCase):
    def setUp(self):
        super(TestExampleWithExtendedSequencedItemType, self).setUp()
        self.datastore.setup_connection()
        self.datastore.setup_tables()

    def tearDown(self):
        self.datastore.drop_tables()
        self.datastore.close_connection()
        super(TestExampleWithExtendedSequencedItemType, self).setUp()

    def construct_datastore(self):
        return SQLAlchemyDatastore(
            base=Base,
            settings=SQLAlchemySettings(),
            tables=(ExtendedIntegerSequencedRecord,)
        )

    def test(self):
        with ExampleApplicationWithExtendedSequencedItemType(self.datastore.session) as app:
            # Create entity.
            entity1 = create_new_example(a='a', b='b')
            self.assertIsInstance(entity1.id, UUID)
            self.assertEqual(entity1.a, 'a')
            self.assertEqual(entity1.b, 'b')

            # Check there is a stored event.
            all_records = list(app.event_store.record_manager.get_notifications())
            self.assertEqual(len(all_records), 1)
            record = all_records[0]
            self.assertEqual(record.sequence_id, entity1.id)
            self.assertEqual(record.position, 0)
            self.assertEqual(record.event_type, 'Example.Created', record.event_type)
            self.assertEqual(record.timestamp, entity1.__created_on__)

            # Read entity from repo.
            retrieved_obj = app.repository[entity1.id]
            self.assertEqual(retrieved_obj.id, entity1.id)



from eventsourcing.infrastructure.base import DEFAULT_PIPELINE_ID
from eventsourcing.infrastructure.datastore import AbstractDatastore
from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.infrastructure.sequenceditem import SequencedItem
from eventsourcing.infrastructure.sequenceditemmapper import AbstractSequencedItemMapper, SequencedItemMapper


class InfrastructureFactory(object):
    """
    Base class for infrastructure factories.
    """
    record_manager_class = None
    sequenced_item_class = SequencedItem
    sequenced_item_mapper_class = SequencedItemMapper
    integer_sequenced_record_class = None
    integer_sequenced_noid_record_class = None
    timestamp_sequenced_record_class = None
    snapshot_record_class = None
    json_encoder_class = None
    json_decoder_class = None
    event_store_class = EventStore

    def __init__(self, record_manager_class=None, sequenced_item_class=None, event_store_class=None,
                 sequenced_item_mapper_class=None, json_encoder_class=None, json_decoder_class=None,
                 integer_sequenced_record_class=None, timestamp_sequenced_record_class=None,
                 snapshot_record_class=None, contiguous_record_ids=False,
                 application_name=None, pipeline_id=DEFAULT_PIPELINE_ID):
        self.record_manager_class = record_manager_class or type(self).record_manager_class
        self.event_store_class = event_store_class or type(self).event_store_class
        self.sequenced_item_class = sequenced_item_class or type(self).sequenced_item_class
        self.sequenced_item_mapper_class = sequenced_item_mapper_class or type(self).sequenced_item_mapper_class
        self.json_encoder_class = json_encoder_class or type(self).json_encoder_class
        self.json_decoder_class = json_decoder_class or type(self).json_decoder_class

        self._integer_sequenced_record_class = integer_sequenced_record_class

        self._timestamp_sequenced_record_class = timestamp_sequenced_record_class

        self._snapshot_record_class = snapshot_record_class

        self.contiguous_record_ids = contiguous_record_ids
        self.application_name = application_name
        self.pipeline_id = pipeline_id

    def construct_integer_sequenced_record_manager(self, **kwargs):
        """
        Constructs an integer sequenced record manager.
        """
        integer_sequenced_record_class = self._integer_sequenced_record_class \
                                         or self.integer_sequenced_record_class

        return self.construct_record_manager(
            integer_sequenced_record_class,
            **kwargs
        )

    def construct_timestamp_sequenced_record_manager(self, **kwargs):
        """
        Constructs a timestamp sequenced record manager.
        """
        timestamp_sequenced_record_class = self._timestamp_sequenced_record_class \
                                           or self.timestamp_sequenced_record_class

        return self.construct_record_manager(
            timestamp_sequenced_record_class,
            **kwargs
        )

    def construct_snapshot_record_manager(self, **kwargs):
        """
        Constructs a snapshot record manager.
        """
        snapshot_record_class = self._snapshot_record_class or self.snapshot_record_class

        return self.construct_record_manager(
            snapshot_record_class,
            **kwargs
        )

    def construct_record_manager(self, record_class, sequenced_item_class=None, **kwargs):
        """
        Constructs an record manager.
        """
        return self.record_manager_class(
            sequenced_item_class=sequenced_item_class or self.sequenced_item_class,
            record_class=record_class,
            contiguous_record_ids=self.contiguous_record_ids,
            application_name=self.application_name,
            pipeline_id=self.pipeline_id,
            **kwargs
        )

    def construct_sequenced_item_mapper(self, cipher):
        """
        Constructs sequenced item mapper object.

        :returns: Sequenced item mapper object.
        :rtype: AbstractSequencedItemMapper
        """
        return self.sequenced_item_mapper_class(
            sequenced_item_class=self.sequenced_item_class,
            cipher=cipher,
            # sequence_id_attr_name=sequence_id_attr_name,
            # position_attr_name=position_attr_name,
            json_encoder_class=self.json_encoder_class,
            json_decoder_class=self.json_decoder_class,
        )

    def construct_integer_sequenced_event_store(self, cipher):
        """
        Constructs an integer sequenced event store.
        """
        sequenced_item_mapper = self.construct_sequenced_item_mapper(cipher)
        record_manager = self.construct_integer_sequenced_record_manager()
        return self.event_store_class(
            record_manager,
            sequenced_item_mapper=sequenced_item_mapper,
        )

    def construct_datastore(self):
        """
        Constructs datastore object.

        :returns: Concrete datastore object object.
        :rtype: AbstractDatastore
        """
        return None



from eventsourcing.tests.sequenced_item_tests.test_django_record_manager import DjangoTestCase
from eventsourcing.application.django import DjangoApplication
from eventsourcing.tests.test_process import TestProcessApplication


class TestProcessWithDjango(DjangoTestCase, TestProcessApplication):
    process_class = DjangoApplication


del TestProcessApplication



# # coding=utf-8
# from __future__ import unicode_literals
#
# from collections import OrderedDict
# from time import sleep
# from uuid import uuid4
#
# from eventsourcing.domain.model.collection import Collection
# from eventsourcing.domain.model.entity import AbstractEntityRepository, AttributeChanged, Created, Discarded, \
#     TimestampedVersionedEntity, attribute, mutate_entity
# from eventsourcing.domain.model.events import TimestampedVersionedEntityEvent, publish, mutator
# from eventsourcing.exceptions import ConcurrencyError, RepositoryKeyError
#
# # Use a private code point to terminate the strings.
# STRING_ID_END = '\uEFFF'
#
#
# class GeneralizedSuffixTree(TimestampedVersionedEntity):
#     """A suffix tree for string matching. Uses Ukkonen's algorithm
#     for construction.
#
#     Adding strings to individual suffix trees needs to be serialized,
#     so you need to use locks around calls to add_string() in a
#     distributed application.
#     """
#
#     class Created(Created):
#         pass
#
#     class AttributeChanged(AttributeChanged):
#         pass
#
#     class Discarded(Discarded):
#         pass
#
#     def __init__(self, root_node_id, case_insensitive=False, **kwargs):
#         super(GeneralizedSuffixTree, self).__init__(**kwargs)
#         self._root_node_id = root_node_id
#         self._case_insensitive = case_insensitive
#         # self._nodes = {}
#         self._string = None
#         self._suffix = None
#         self._node_repo = None
#         self._node_child_collection_repo = None
#         self._edge_repo = None
#         self._stringid_collection_repo = None
#
#     @attribute
#     def string(self):
#         return self._string
#
#     @property
#     def nodes(self):
#         return self._nodes
#
#     @property
#     def root_node_id(self):
#         return self._root_node_id
#
#     @property
#     def suffix(self):
#         return self._suffix
#
#     @property
#     def case_insensitive(self):
#         return self._case_insensitive
#
#     def __repr__(self):
#         """
#         Lists edges in the suffix tree
#         """
#         # Todo: Something that shows the actual tree.
#         return ('GeneralizedSuffixTree(id={0.id}, root_node_id={0.root_node_id}, case_insensitive={'
#                 '0.case_insensitive})'.format(self))
#
#     def add_string(self, string, string_id):
#         assert isinstance(string, str)
#         assert isinstance(string_id, str)
#
#         # Lower the string, if it's a case insensitive tree.
#         if self._case_insensitive:
#             string = string.lower()
#         assert STRING_ID_END not in string_id
#         assert STRING_ID_END not in string
#
#         # Append the (hopefully unique) string ID to the string.
#         string_extended = u"{}{}".format(string, STRING_ID_END)
#
#         self._suffix = Suffix(self._root_node_id, 0, -1)
#         for i in range(len(string_extended)):
#             self._add_prefix(i, string_extended, string_id)
#
#     def _add_prefix(self, i, string, string_id):
#         """The core construction method.
#         """
#         last_parent_node_id = None
#         parent_node_id = None
#         added_node_ids = set()
#         while True:
#
#             parent_node_id = self.suffix.source_node_id
#
#             # Break if prefix is already in tree.
#             if self.suffix.explicit():
#                 # Explicit suffix points to an existing node. Try to find the node.
#                 prefix = string[i]
#                 try:
#                     edge = self.get_edge(make_edge_id(parent_node_id, prefix))
#                 except KeyError:
#                     # If the edge doesn't exist, the prefix isn't in the
#                     # tree, so continue by adding an edge and a node.
#                     # print("Explicit suffix, edge not found for letter '{}'".format(prefix))
#                     pass
#                 else:
#                     # The edge was found, so exit the loop.
#                     # print("Explicit suffix, edge found for letter '{}': {}".format(prefix, edge))
#
#                     # Although, might need to add string ID to node collection, if it's a leaf and exact match.
#                     dest_node = self.get_node(edge.dest_node_id)
#                     if self.is_leaf(dest_node):
#
#                         # If it's an exact match...
#                         if edge.label == string[i:]:
#
#                             # Add the string ID to the node collection.
#                             self.add_string_id(edge.dest_node_id, string_id)
#                             # print("Added string ID '{}' to explicit exact match {}".format(string_id, dest_node))
#
#                             # Attempt to replicate for a duplicate what happens first time a string was added.
#                             if edge.dest_node_id not in added_node_ids:
#                                 # Adjust the active suffix.
#                                 self.follow_suffix_link_or_increment_suffix_first_char(string)
#
#                     break
#
#             else:
#                 # Break if prefix is already in tree.
#                 prefix = string[self.suffix.first_char_index]
#                 edge = self.get_edge(make_edge_id(self.suffix.source_node_id, prefix))
#                 suffix_label_offset = self.suffix.length + 1
#                 if suffix_label_offset > len(edge.label) - 1:
#                     raise ValueError(
#                         "Suffix offset too large for label: %s: %s" % (suffix_label_offset, edge.label[:100]))
#                 # print("Implicit edge dest node: {}".format(edge.dest_node_id))
#                 if edge.label[suffix_label_offset] == string[i]:
#                     # print("Matched edge label til end of suffix: {} {}".format(edge.label[:100],
#                     # suffix_label_offset))
#                     # The suffix was found in this edge, so exit the loop.
#
#                     # Add the string ID to the node collection.
#                     self.add_string_id(edge.dest_node_id, string_id)
#                     # print("Added string ID '{}' to implicit exact match: {}".format(string_id, self.get_node(
#                     # edge.dest_node_id)))
#
#                     break
#
#                 else:
#                     # Split the edge, with a new middle node that will be the parent node.
#                     parent_node_id = self._split_edge(edge, self.suffix)
#
#             # Otherwise the prefix isn't in the tree, so create a node and an edge for this prefix.
#             new_leaf_node = register_new_node(string_id=string_id, is_leaf=True)
#
#             # Remeber this node was created in this call.
#             added_node_ids.add(new_leaf_node.id)
#
#             # Make a new edge from the parent node to the
#             # new leaf node. Make a label for the edge to
#             # the leaf, using the suffix of the string
#             # from i onwards.
#             new_edge = register_new_edge(
#                 edge_id=make_edge_id(parent_node_id, string[i]),
#                 label=string[i:],
#                 source_node_id=parent_node_id,
#                 dest_node_id=new_leaf_node.id,
#             )
#
#             # print("For letter '{}' added node and edge: {}".format(string[i], new_edge))
#
#             # Add the new leaf node as a child node of the parent node.
#             self.add_child_node(parent_node_id, new_leaf_node.id, new_edge.length + 1)
#
#             # Add the string ID to the node collection.
#             self.add_string_id(new_leaf_node.id, string_id)
#             # print("Added string ID '{}' to new node {}".format(string_id, new_leaf_node))
#
#             # Create suffix link from last parent to current parent (unless it's root).
#             self.create_suffix_link(last_parent_node_id, parent_node_id)
#
#             # Remember the last parent node.
#             last_parent_node_id = parent_node_id
#
#             # Adjust the active suffix.
#             self.follow_suffix_link_or_increment_suffix_first_char(string)
#
#         # Create suffix link from last parent to current parent (unless it's root).
#         self.create_suffix_link(last_parent_node_id, parent_node_id)
#
#         # Which is the last leaf node?
#
#         self.suffix.last_char_index += 1
#         # print("Incremented suffix last char index: {}".format(self.suffix.last_char_index))
#
#         self._canonize_suffix(self.suffix, string)
#         # print("Cannonized suffix: {}".format(self.suffix))
#
#     def _split_edge(self, first_edge, suffix):
#         assert isinstance(first_edge, SuffixTreeEdge)
#         assert isinstance(suffix, Suffix)
#
#         # Split the label.
#         orig_label = first_edge.label
#         first_label = orig_label[:suffix.length + 1]
#         second_label = orig_label[suffix.length + 1:]
#
#         # print("Splitting edge: {}  <==>  {}".format(first_label[:100], second_label[:100]))
#
#         # Create a new middle node that will split the edge.
#         new_middle_node = register_new_node(suffix_node_id=suffix.source_node_id)
#         # self._cache_node(new_middle_node)
#
#         # print("New middle node: {}".format(new_middle_node))
#
#         # Create a new edge, from the new middle node to the original destination.
#         second_edge_id = make_edge_id(source_node_id=new_middle_node.id, first_char=second_label[0])
#         original_dest_node_id = first_edge.dest_node_id
#
#         second_edge = register_new_edge(
#             label=second_label,
#             edge_id=second_edge_id,
#             source_node_id=new_middle_node.id,
#             dest_node_id=original_dest_node_id,
#         )
#
#         # print("New second edge: {}".format(second_edge))
#
#         # Add the original dest node as a child of the new middle node.
#         self.add_child_node(
#             parent_node_id=new_middle_node.id,
#             child_node_id=original_dest_node_id,
#             edge_length=second_edge.length + 1,
#         )
#
#         # Shorten the first edge, so its destination is the new middle node.
#         try:
#             first_edge.shorten(label=first_label, dest_node_id=new_middle_node.id)
#             # print("Shortened first edge: {}".format(first_edge))
#         except ConcurrencyError as e:
#             # If the first edge has changed by now, then abort this split by
#             # discarding the new_middle_node, which hasn't been connected to
#             # the tree yet, so can safely be discarded.
#             new_middle_node.discard()
#             sleep(0.1)
#             raise e
#
#         # Remove the original dest node as child of the original
#         # source node, and add the new middle node instead.
#         self.switch_child_node(
#             parent_node_id=first_edge.source_node_id,
#             old_child_id=original_dest_node_id,
#             new_child_id=new_middle_node.id,
#             edge_length=first_edge.length + 1,
#         )
#
#         # Return middle node.
#         return new_middle_node.id
#
#     def add_child_node(self, parent_node_id, child_node_id, edge_length):
#         collection = self.get_or_create_node_child_collection(parent_node_id)
#         if child_node_id not in collection.child_node_ids:
#             # - also pass in the edge length (used when finding strings)
#             collection.add_child(child_node_id, edge_length)
#
#     def switch_child_node(self, parent_node_id, old_child_id, new_child_id, edge_length):
#         collection = self.get_or_create_node_child_collection(parent_node_id)
#         assert isinstance(collection, SuffixTreeNodeChildCollection)
#         collection.switch_child(old_child_id, new_child_id, edge_length)
#
#     def add_string_id(self, parent_node_id, string_id):
#         collection = self.get_or_create_stringid_collection(parent_node_id)
#         if string_id not in collection.items:
#             collection.add_item(string_id)
#
#     def create_suffix_link(self, last_parent_node_id, parent_node_id):
#         if last_parent_node_id is not None:
#             assert parent_node_id is not None
#
#             # Set suffix_node_id on last_parent_node to parent_node_id.
#             last_parent_node = self.get_node(last_parent_node_id)
#             assert isinstance(last_parent_node, SuffixTreeNode)
#             last_parent_node.suffix_node_id = parent_node_id
#
#     def follow_suffix_link_or_increment_suffix_first_char(self, string):
#         # If we're at root, there isn't a suffix link, so increment the first char index.
#         if self.suffix.source_node_id == self.root_node_id:
#             # print("Incrementing suffix first char...")
#             self.suffix.first_char_index += 1
#         else:
#             # If we're not at root, follow the suffix link.
#             last_source_node_id = self.suffix.source_node_id
#             next_source_node_id = self.get_node(last_source_node_id).suffix_node_id
#             # print("Following suffix link from {} to {}".format(last_source_node_id, next_source_node_id))
#             self.suffix.source_node_id = next_source_node_id
#         self._canonize_suffix(self.suffix, string)
#
#     def is_leaf(self, dest_node):
#         return dest_node.is_leaf
#
#     def is_root_node(self, dest_node):
#         return dest_node.is_root
#
#     def remove_string(self, string, string_id):
#         # Todo: Maybe this is an application service?
#
#         assert isinstance(string_id, str)
#         if self._case_insensitive:
#             string = string.lower()
#         assert STRING_ID_END not in string_id
#         assert STRING_ID_END not in string
#         # string += u"{}{}".format(string_id, STRING_ID_END)
#         string += u"{}".format(STRING_ID_END)
#
#         # Disassociate the string ID from its leaf nodes.
#         #  - but don't discard the node entity, because we might
#         #    need to add this suffix back into the tree,
#         #    and a hard prune would be relatively complicated.
#         #  - although, it might be useful to have something do garbage collection
#         from eventsourcing.contrib.suffixtrees.domain.services.generalizedsuffixtree import find_substring_edge
#
#         for i in range(len(string)):
#
#             # Find the suffix in the tree.
#
#             suffix = string[i:]
#             edge, ln = find_substring_edge(suffix, self, self._edge_repo)
#
#             if edge is None:
#                 continue
#
#             # Get the leaf node.
#             leaf_node = self.get_node(edge.dest_node_id)
#             assert isinstance(leaf_node, SuffixTreeNode)
#             if leaf_node.string_id == string_id:
#                 leaf_node.string_id = None
#             try:
#                 stringid_collection = self._stringid_collection_repo[leaf_node.id]
#             except KeyError:
#                 pass
#             else:
#                 assert isinstance(stringid_collection, StringidCollection)
#                 if string_id in stringid_collection.items:
#                     stringid_collection.remove_item(string_id)
#
#     def get_edge(self, edge_id):
#         # Raises KeyError is not found because this method was factored out from various places that
#         # used the repo directly and so expected a KeyError to be raised if edge is not in repo.
#         return self._edge_repo[edge_id]
#
#     def has_edge(self, edge_id):
#         try:
#             return self.get_edge(edge_id)
#         except KeyError:
#             pass
#
#     def get_node(self, node_id):
#         # Raises KeyError is not found because this method was factored out from various places that
#         # used the repo directly and so expected a KeyError to be raised if node is not in repo.
#         # Todo: Change this (and callers) to return (and handle) None when the edge is not found.
#         """
#
#         :rtype: SuffixTreeNode
#         """
#         # if not self._node_repo:
#         #     return self.nodes[node_id]
#         # try:
#         #     return self.nodes[node_id]
#         # except KeyError as e:
#         #     try:
#         #         node = self._node_repo[node_id]
#         #     except KeyError:
#         #         raise e
#         #     else:
#         #         self._cache_node(node)
#         #         return node
#
#         return self._node_repo[node_id]
#
#     def _cache_node(self, node):
#         self.nodes[node.id] = node
#
#     def get_or_create_node_child_collection(self, node_id):
#         """
#         Returns children of node.
#
#         :rtype: SuffixTreeNodeChildCollection
#         """
#         try:
#             node = self._node_child_collection_repo[node_id]
#         except RepositoryKeyError:
#             node = register_new_node_child_collection(node_id)
#         return node
#
#     def get_or_create_stringid_collection(self, node_id):
#         """
#         Returns collection of string IDs.
#
#         :rtype: StringidCollection
#         """
#         try:
#             node = self._stringid_collection_repo[node_id]
#         except KeyError:
#             node = register_new_stringid_collection(node_id)
#         return node
#
#     def _canonize_suffix(self, suffix, string):
#         """This canonizes the suffix, walking along its suffix string until it
#         is explicit or there are no more matched nodes.
#         """
#         if not suffix.explicit():
#             if len(string) < suffix.first_char_index + 1:
#                 raise AssertionError("String length is {} and suffix first char index is {}"
#                                      "".format(len(string), suffix.first_char_index))
#             edge_id = make_edge_id(suffix.source_node_id, string[suffix.first_char_index])
#             e = self.get_edge(edge_id)
#             if e.length <= suffix.length:
#                 suffix.first_char_index += e.length + 1
#                 suffix.source_node_id = e.dest_node_id
#                 self._canonize_suffix(suffix, string)
#
#
# class SuffixTreeNode(TimestampedVersionedEntity):
#     """A node in the suffix tree.
#     """
#
#     class Created(Created): pass
#
#     class AttributeChanged(AttributeChanged): pass
#
#     class Discarded(Discarded): pass
#
#     def __init__(self, suffix_node_id=None, string_id=None, is_leaf=False, is_root=False, *args, **kwargs):
#         super(SuffixTreeNode, self).__init__(*args, **kwargs)
#         self._suffix_node_id = suffix_node_id
#         self._string_id = string_id
#         self._is_leaf = is_leaf
#         self._is_root = is_root
#
#     @attribute
#     def string_id(self):
#         """The id of a string being added to the generalised suffix tree when this node was created.
#         """
#
#     @attribute
#     def suffix_node_id(self):
#         """The ID of the node representing the suffix of the suffix represented by this node.
#         """
#
#     @property
#     def is_leaf(self):
#         return self._is_leaf
#
#     @property
#     def is_root(self):
#         return self._is_root
#
#     def __repr__(self):
#         return "Node(id='{}')".format(self.id)
#
#
# class StringidCollection(Collection):
#     class Created(Collection.Created):
#         pass
#
#     class ItemAdded(Collection.ItemAdded):
#         pass
#
#     class ItemRemoved(Collection.ItemRemoved):
#         pass
#
#
# class SuffixTreeNodeChildCollection(TimestampedVersionedEntity):
#     """A collecton of child nodes in the suffix tree.
#     """
#
#     class Created(Created): pass
#
#     class AttributeChanged(AttributeChanged): pass
#
#     class Discarded(Discarded): pass
#
#     class ChildNodeAdded(TimestampedVersionedEntityEvent): pass
#
#     class ChildNodeRemoved(TimestampedVersionedEntityEvent): pass
#
#     class ChildNodeSwitched(TimestampedVersionedEntityEvent): pass
#
#     def __init__(self, *args, **kwargs):
#         super(SuffixTreeNodeChildCollection, self).__init__(*args, **kwargs)
#         self._child_node_ids = OrderedDict()
#
#     @property
#     def child_node_ids(self):
#         return self._child_node_ids
#
#     def add_child(self, child_node_id, edge_len):
#         event = SuffixTreeNodeChildCollection.ChildNodeAdded(
#             originator_id=self.id,
#             originator_version=self.__version__,
#             child_node_id=child_node_id,
#             edge_len=edge_len,
#         )
#         self._apply(event)
#         publish(event)
#
#     def switch_child(self, old_node_id, new_node_id, new_edge_len):
#         event = SuffixTreeNodeChildCollection.ChildNodeSwitched(
#             originator_id=self.id,
#             originator_version=self.__version__,
#             old_node_id=old_node_id,
#             new_node_id=new_node_id,
#             new_edge_len=new_edge_len,
#         )
#         self._apply(event)
#         publish(event)
#
#     @staticmethod
#     def _mutator(initial, event):
#         return suffix_tree_node_child_collection_mutator(initial, event)
#
#
# @mutator
# def suffix_tree_node_child_collection_mutator(initial, event):
#     return mutate_entity(initial, event)
#
#
# @suffix_tree_node_child_collection_mutator.register(SuffixTreeNodeChildCollection.ChildNodeAdded)
# def child_node_child_collection_added_mutator(self, event):
#     assert isinstance(self, SuffixTreeNodeChildCollection), self
#     self._child_node_ids[event.child_node_id] = event.edge_len
#     self._increment_version()
#     return self
#
#
# @suffix_tree_node_child_collection_mutator.register(SuffixTreeNodeChildCollection.ChildNodeSwitched)
# def child_node_child_collection_switched_mutator(self, event):
#     assert isinstance(self, SuffixTreeNodeChildCollection), self
#     try:
#         del (self._child_node_ids[event.old_node_id])
#     except KeyError:
#         pass
#     self._child_node_ids[event.new_node_id] = event.new_edge_len
#     self._increment_version()
#     return self
#
#
# class SuffixTreeEdge(TimestampedVersionedEntity):
#     """An edge in the suffix tree.
#     """
#
#     class Created(Created):
#         pass
#
#     class AttributeChanged(AttributeChanged):
#         pass
#
#     class Shortened(TimestampedVersionedEntityEvent):
#         @property
#         def label(self):
#             return self.__dict__['label']
#
#         @property
#         def dest_node_id(self):
#             return self.__dict__['dest_node_id']
#
#     class Discarded(Discarded):
#         pass
#
#     def __init__(self, label, source_node_id, dest_node_id, **kwargs):
#         super(SuffixTreeEdge, self).__init__(**kwargs)
#         self._label = label
#         self._source_node_id = source_node_id
#         self._dest_node_id = dest_node_id
#
#     @attribute
#     def label(self):
#         """Index of start of string part represented by this edge.
#         """
#         return self._label
#
#     @attribute
#     def source_node_id(self):
#         """Id of source node of edge.
#         """
#         return self._source_node_id
#
#     @attribute
#     def dest_node_id(self):
#         """Id of destination node of edge.
#         """
#         return self._dest_node_id
#
#     @property
#     def length(self):
#         """Number of chars in the string part represented by this edge.
#         """
#         return len(self.label) - 1
#
#     def __repr__(self):
#         return ("Edge(id='{}' source='{}' dest='{}', label='{}')"
#                 "".format(self.id, self.source_node_id, self.dest_node_id, self.label[:30]))
#
#     def shorten(self, label, dest_node_id):
#         self._assert_not_discarded()
#         event = SuffixTreeEdge.Shortened(
#             originator_id=self._id,
#             originator_version=self.___version__,
#             label=label,
#             dest_node_id=dest_node_id,
#         )
#         self._apply(event)
#         publish(event)
#
#     @staticmethod
#     def _mutator(initial, event):
#         return suffix_tree_edge_mutator(initial, event)
#
#
# @mutator
# def suffix_tree_edge_mutator(initial, event):
#     return mutate_entity(initial, event)
#
#
# @suffix_tree_edge_mutator.register(SuffixTreeEdge.Shortened)
# def suffix_tree_edge_shortened_mutator(self, event):
#     assert isinstance(event, SuffixTreeEdge.Shortened), event
#     assert isinstance(self, SuffixTreeEdge), type(self)
#     self._validate_originator(event)
#     self._label = event.label
#     self._dest_node_id = event.dest_node_id
#     self._increment_version()
#     return self
#
#
# class Suffix(object):
#     """Represents a suffix from first_char_index to last_char_index.
#     """
#
#     def __init__(self, source_node_id, first_char_index, last_char_index):
#         """
#         :type source_node_id: str
#             Index of node where this suffix starts.
#
#         :type first_char_index: str
#             Index of start of suffix in string.
#
#         :type last_char_index: str
#             Index of end of suffix in string.
#         """
#         assert isinstance(source_node_id, str)
#         assert isinstance(first_char_index, int)
#         assert isinstance(last_char_index, int)
#         assert source_node_id is not None
#         self.source_node_id = source_node_id
#         self.first_char_index = first_char_index
#         self.last_char_index = last_char_index
#
#     @property
#     def length(self):
#         """Number of chars in string.
#         """
#         return self.last_char_index - self.first_char_index
#
#     def explicit(self):
#         """A suffix is explicit if it ends on a node. first_char_index
#         is set greater than last_char_index to indicate this.
#         """
#         return self.first_char_index > self.last_char_index
#
#     def implicit(self):
#         return self.last_char_index >= self.first_char_index
#
#     def __repr__(self):
#         return "Suffix(implicit={} first={}, last={}, node={})".format(self.implicit(), self.first_char_index,
#                                                                        self.last_char_index, self.source_node_id)
#
#
# # Factory methods.
#
# def register_new_node(suffix_node_id=None, string_id=None, is_leaf=False, is_root=False):
#     """Factory method, registers new node.
#     """
#     # Make a unique ID.
#     node_id = uuid4().hex
#
#     # Instantiate a Created event.
#     created_event = SuffixTreeNode.Created(
#         originator_id=node_id,
#         suffix_node_id=suffix_node_id,
#         string_id=string_id,
#         is_leaf=is_leaf,
#         is_root=is_root,
#     )
#
#     # Make the new Node entity.
#     new_node = SuffixTreeNode.mutate(event=created_event)
#
#     # Publish the Created event.
#     publish(created_event)
#
#     # Also create a new child collection, to
#     # avoid any contention around this event.
#     register_new_node_child_collection(node_id)
#
#     # Return the new node.
#     return new_node
#
#
# def register_new_node_child_collection(node_id):
#     """Factory method, registers new node child collection.
#     """
#     event = SuffixTreeNodeChildCollection.Created(
#         originator_id=node_id,
#     )
#     entity = SuffixTreeNodeChildCollection.mutate(event=event)
#     publish(event)
#     return entity
#
#
# def register_new_stringid_collection(node_id):
#     """Factory method, registers new collection of string IDs.
#     """
#     event = StringidCollection.Created(
#         originator_id=node_id,
#     )
#     entity = StringidCollection.mutate(event=event)
#     publish(event)
#     return entity
#
#
# def make_edge_id(source_node_id, first_char):
#     """Returns a string made from given params.
#     """
#     # assert STRING_ID_PADDING_RIGHT not in first_char
#     # assert STRING_ID_PADDING_RIGHT not in source_node_id
#     return "{}::{}".format(source_node_id, first_char)
#
#
# def register_new_edge(edge_id, label, source_node_id, dest_node_id):
#     """Factory method, registers new edge.
#     """
#     event = SuffixTreeEdge.Created(
#         originator_id=edge_id,
#         label=label,
#         source_node_id=source_node_id,
#         dest_node_id=dest_node_id,
#     )
#     entity = SuffixTreeEdge.mutate(event=event)
#     publish(event)
#     return entity
#
#
# def register_new_suffix_tree(case_insensitive=False):
#     """Factory method, returns new suffix tree object.
#     """
#     assert isinstance(case_insensitive, bool)
#
#     suffix_tree_id = uuid4().hex
#
#     root_node = register_new_node(is_root=True)
#     assert isinstance(root_node, SuffixTreeNode)
#
#     event = GeneralizedSuffixTree.Created(
#         originator_id=suffix_tree_id,
#         root_node_id=root_node.id,
#         case_insensitive=case_insensitive,
#     )
#     entity = GeneralizedSuffixTree.mutate(event=event)
#
#     assert isinstance(entity, GeneralizedSuffixTree)
#
#     # entity.nodes[root_node.id] = root_node
#
#     publish(event)
#
#     return entity
#
#
# # Repositories.
#
# class GeneralizedSuffixTreeRepository(AbstractEntityRepository):
#     pass
#
#
# class NodeRepository(AbstractEntityRepository):
#     pass
#
#
# class EdgeRepository(AbstractEntityRepository):
#     pass



from django.db import IntegrityError, connection, transaction, ProgrammingError

from eventsourcing.infrastructure.base import SQLRecordManager


class DjangoRecordManager(SQLRecordManager):

    _where_application_name_tmpl = (
        " WHERE application_name = %s AND pipeline_id = %s"
    )

    def write_records(self, records, tracking_kwargs=None, orm_objs=None):
        try:
            with transaction.atomic(self.record_class.objects.db):
                with connection.cursor() as cursor:
                    # Insert tracking record.
                    if tracking_kwargs:
                        params = [tracking_kwargs[c] for c in self.tracking_record_field_names]
                        cursor.execute(self.insert_tracking_record, params)

                    if self.contiguous_record_ids:
                        # Use cursor to execute insert select max statement.
                        for record in records:
                            # Get values from record obj.
                            # List of params, because dict doesn't work with Django and SQLite.
                            params = []

                            for col_name in self.field_names:
                                col_value = getattr(record, col_name)
                                col_type = self.record_class._meta.get_field(col_name)

                                # Prepare value for database.
                                param = col_type.get_db_prep_value(col_value, connection)
                                params.append(param)

                            # Notification logs fields, to be inserted with event fields.
                            if hasattr(self.record_class, 'application_name'):
                                params.append(self.application_name)
                            if hasattr(self.record_class, 'pipeline_id'):
                                params.append(self.pipeline_id)
                            if hasattr(record, 'causal_dependencies'):
                                params.append(record.causal_dependencies)

                            # Where clause fields.
                            if hasattr(self.record_class, 'application_name'):
                                params.append(self.application_name)
                            if hasattr(self.record_class, 'pipeline_id'):
                                params.append(self.pipeline_id)

                            # Execute insert statement.
                            cursor.execute(self.insert_select_max, params)
                            # Todo: Use insert_values when records have IDs (like SQLAlchemy manager).
                            # Todo: Support 'event-not-notifiable' by setting pipeline ID and notification ID to None.

                    else:
                        # This can only work for simple models, without application_name
                        # and pipeline_id, because it relies on the auto-incrementing ID.
                        # Todo: If it's faster, change to use an "insert_values" raw query.
                        # Save record objects.
                        for record in records:
                            record.save()

                    # Call 'save()' on each of the ORM objects.
                    if orm_objs:
                        for orm_obj in orm_objs:
                            orm_obj.save()

        except IntegrityError as e:
            self.raise_record_integrity_error(e)

    def _prepare_insert(self, tmpl, record_class, field_names, placeholder_for_id=False):
        """
        With transaction isolation level of "read committed" this should
        generate records with a contiguous sequence of integer IDs, using
        an indexed ID column, the database-side SQL max function, the
        insert-select-from form, and optimistic concurrency control.
        """
        field_names = list(field_names)
        if hasattr(record_class, 'application_name') and 'application_name' not in field_names:
            field_names.append('application_name')
        if hasattr(record_class, 'pipeline_id') and 'pipeline_id' not in field_names:
            field_names.append('pipeline_id')
        if hasattr(record_class, 'causal_dependencies') and 'causal_dependencies' not in field_names:
            field_names.append('causal_dependencies')
        if placeholder_for_id:
            if self.notification_id_name:
                if self.notification_id_name not in field_names:
                    field_names.append('id')

        statement = tmpl.format(
            tablename=self.get_record_table_name(record_class),
            columns=", ".join(field_names),
            placeholders=", ".join(['%s' for _ in field_names]),
            notification_id=self.notification_id_name
        )
        return statement

    def get_record_table_name(self, record_class):
        """Returns table name from SQLAlchemy record class."""
        return record_class._meta.db_table

    def get_record(self, sequence_id, position):
        kwargs = {
            self.field_names.sequence_id: sequence_id,
            self.field_names.position: position,
        }
        records = self.record_class.objects.filter(**kwargs)
        # Todo: try/except for native error here, call self.raise_index_error()
        return records.all()[0]

    def get_records(self, sequence_id, gt=None, gte=None, lt=None, lte=None, limit=None,
                    query_ascending=True, results_ascending=True):

        assert limit is None or limit >= 1, limit

        filter_kwargs = {self.field_names.sequence_id: sequence_id}
        objects = self.record_class.objects.filter(**filter_kwargs)

        if hasattr(self.record_class, 'application_name'):
            objects = objects.filter(application_name=self.application_name)

        position_field_name = self.field_names.position

        if query_ascending:
            objects = objects.order_by(position_field_name)
        else:
            objects = objects.order_by('-' + position_field_name)

        if gt is not None:
            arg = '{}__gt'.format(position_field_name)
            objects = objects.filter(**{arg: gt})
        if gte is not None:
            arg = '{}__gte'.format(position_field_name)
            objects = objects.filter(**{arg: gte})
        if lt is not None:
            arg = '{}__lt'.format(position_field_name)
            objects = objects.filter(**{arg: lt})
        if lte is not None:
            arg = '{}__lte'.format(position_field_name)
            objects = objects.filter(**{arg: lte})

        if limit is not None:
            objects = objects[:limit]

        records = objects.all()

        if results_ascending != query_ascending:
            # This code path is under test, but not otherwise used ATM.
            records = list(records)
            records.reverse()

        return records

    def get_notifications(self, start=None, stop=None, *args, **kwargs):
        """
        Returns all records in the table.
        """
        filter_kwargs = {}
        # Todo: Also support sequencing by 'position' if items are sequenced by timestamp?
        if start is not None:
            filter_kwargs['%s__gte' % self.notification_id_name] = start + 1
        if stop is not None:
            filter_kwargs['%s__lt' % self.notification_id_name] = stop + 1
        objects = self.record_class.objects.filter(**filter_kwargs)

        if hasattr(self.record_class, 'application_name'):
            objects = objects.filter(application_name=self.application_name)
        if hasattr(self.record_class, 'pipeline_id'):
            objects = objects.filter(pipeline_id=self.pipeline_id)

        objects = objects.order_by('%s' % self.notification_id_name)
        return objects.all()

    def delete_record(self, record):
        """
        Permanently removes record from table.
        """
        record.delete()

    def get_max_record_id(self):
        assert self.notification_id_name
        try:
            objects = self.record_class.objects
            if hasattr(self.record_class, 'application_name'):
                objects = objects.filter(application_name=self.application_name)
            if hasattr(self.record_class, 'pipeline_id'):
                objects = objects.filter(pipeline_id=self.pipeline_id)
            latest = objects.latest(self.notification_id_name)
            return getattr(latest, self.notification_id_name)
        except (self.record_class.DoesNotExist, ProgrammingError):
            return None

    def get_max_tracking_record_id(self, upstream_application_name):
        notification_id = None
        try:
            objects = self.tracking_record_class.objects
            objects = objects.filter(application_name=self.application_name)
            objects = objects.filter(upstream_application_name=upstream_application_name)
            objects = objects.filter(pipeline_id=self.pipeline_id)
            notification_id = objects.latest('notification_id').notification_id
        except self.tracking_record_class.DoesNotExist:
            pass
        return notification_id

    def has_tracking_record(self, upstream_application_name, pipeline_id, notification_id):
        objects = self.tracking_record_class.objects
        objects = objects.filter(application_name=self.application_name)
        objects = objects.filter(upstream_application_name=upstream_application_name)
        objects = objects.filter(pipeline_id=pipeline_id)
        objects = objects.filter(notification_id=notification_id)
        return bool(objects.count())

    def all_sequence_ids(self):
        sequence_id_fieldname = self.field_names.sequence_id
        values_queryset = self.record_class.objects.values(sequence_id_fieldname).distinct()
        for values in values_queryset:
            yield values[sequence_id_fieldname]



from abc import ABC, abstractmethod

from eventsourcing.domain.model.array import BigArray
from eventsourcing.infrastructure.base import ACIDRecordManager

DEFAULT_SECTION_SIZE = 20


class AbstractNotificationLog(ABC):
    """
    Presents a sequence of sections from a sequence of notifications.
    """

    @abstractmethod
    def __getitem__(self, section_id):
        """
        Get section of notification log.

        :rtype: Section
        """


class Section(object):
    """
    Section of a notification log.

    Contains items, and has an ID.

    May also have either IDs of previous and next sections of the notification log.
    """

    def __init__(self, section_id, items, previous_id=None, next_id=None):
        self.section_id = section_id
        self.items = items
        self.previous_id = previous_id
        self.next_id = next_id


class LocalNotificationLog(AbstractNotificationLog):
    """
    Presents a sequence of sections from a sequence of notifications.
    """

    def __init__(self, section_size=None):
        self.section_size = section_size or DEFAULT_SECTION_SIZE
        # self.last_start = None

    def __getitem__(self, section_id):
        # Get section of notification log.
        next_position = None
        if section_id == 'current':

            # Get start of the last section.
            next_position = self.get_next_position()
            start = next_position // self.section_size * self.section_size
        else:

            try:
                first_item_number, _ = section_id.split(',')
            except ValueError as e:
                raise ValueError("Couldn't split '{}': {}".format(section_id, e))

            # Convert from 1-based to 0-based.
            start = int(first_item_number) - 1

            # Go back to nearest section start.
            start = start // self.section_size * self.section_size

        # Get stop index (section doesn't include this one).
        stop = start + self.section_size

        # Get the items.
        items = list(self.get_items(start, stop, next_position))

        # Get previous and next section IDs.
        if start:
            first_item_number = start + 1 - self.section_size
            last_item_number = first_item_number - 1 + self.section_size
            previous_id = self.format_section_id(first_item_number, last_item_number)
        else:
            previous_id = None

        if len(items) == self.section_size:
            first_item_number = start + 1 + self.section_size
            last_item_number = first_item_number - 1 + self.section_size
            next_id = self.format_section_id(first_item_number, last_item_number)
        else:
            next_id = None

        # Return a section of the notification log.
        section_id = self.format_section_id(start + 1, start + self.section_size)
        return Section(
            section_id=section_id,
            items=items,
            previous_id=previous_id,
            next_id=next_id,
        )

    @abstractmethod
    def get_next_position(self):
        """Returns next unoccupied position in zero-based sequence.

        Since the notification IDs are one-based, the next position
        in the zero-based sequence equals the current max notification ID
        which is 1-based. If there are no records, the max notification ID
        will be None, and the next position is zero.

        :returns: Non-negative integer.
        :rtype: int
        """

    @abstractmethod
    def get_items(self, start, stop, next_position=None):
        """
        Returns items for section.

        :rtype: list
        """

    @staticmethod
    def format_section_id(first_item_number, last_item_number):
        return '{},{}'.format(first_item_number, last_item_number)


class RecordManagerNotificationLog(LocalNotificationLog):
    def __init__(self, record_manager, section_size):
        super(RecordManagerNotificationLog, self).__init__(section_size)
        assert isinstance(record_manager, ACIDRecordManager), record_manager
        assert record_manager.contiguous_record_ids
        self.record_manager = record_manager

    def get_items(self, start, stop, next_position=None):
        notifications = []
        for record in self.record_manager.get_notifications(start, stop):
            notification = {
                'id': getattr(record, self.record_manager.notification_id_name)
            }
            for field_name in self.record_manager.field_names:
                notification[field_name] = getattr(record, field_name)
            if hasattr(record, 'causal_dependencies'):
                notification['causal_dependencies'] = record.causal_dependencies
            notifications.append(notification)
        return notifications

    def get_next_position(self):
        """Returns next unoccupied position in zero-based sequence.

        Since the notification IDs are one-based, the next position
        in the zero-based sequence equals the current max notification ID
        which is 1-based. If there are no records, the max notification ID
        will be None, and the next position is zero.

        :returns: Non-negative integer.
        :rtype: int
        """
        return self.record_manager.get_max_record_id() or 0


class BigArrayNotificationLog(LocalNotificationLog):
    def __init__(self, big_array, section_size):
        super(BigArrayNotificationLog, self).__init__(section_size)
        assert isinstance(big_array, BigArray)
        if big_array.repo.array_size % section_size:
            raise ValueError("Section size {} doesn't divide array size {}".format(
                section_size, big_array.repo.array_size
            ))
        self.big_array = big_array

    def get_items(self, start, stop, next_position=None):
        next_position = self.get_next_position() if next_position is None else next_position
        stop = min(stop, next_position)
        return self.big_array[start:stop]

    def get_next_position(self):
        """Returns next unoccupied position in zero-based sequence.

        :returns: Non-negative integer.
        :rtype: int
        """
        return self.big_array.get_next_position()


class NotificationLogReader(ABC):
    def __init__(self, notification_log, use_direct_query_if_available=False):
        assert isinstance(notification_log, AbstractNotificationLog)
        self.notification_log = notification_log
        self.section_count = 0
        self.use_direct_query_if_available = use_direct_query_if_available
        self.position = 0
        self.seek(0)

    def __getitem__(self, item=None):
        if isinstance(item, slice):
            assert item.start >= 0, item.start
            self.seek(item.start)
            return self.read_items(item.stop)
        else:
            assert isinstance(item, int), type(item)
            assert item >= 0, item
            self.seek(item)
            return self.read_list(advance_by=1)[0]

    def __iter__(self):
        return self.read_items()

    def __next__(self):
        try:
            return self.read_list(advance_by=1)[0]
        except IndexError:
            raise StopIteration

    def seek(self, position):
        """
        Sets position of reader in notification log sequence.

        This represents the position of the last notification read by the reader.
        The next notification returned by the reader will be the next position.

        :param int position: Position is notification log sequence.
        :raises ValueError: if the position is less than zero
        """
        if position < 0:
            raise ValueError("Position less than zero: {}".format(position))
        self.position = position

    def read(self, advance_by=None):
        return self.read_items(advance_by=advance_by)

    def read_items(self, stop_index=None, advance_by=None):
        self.section_count = 0

        start_item_num = self.position + 1

        # Validate the position.
        if self.position < 0:
            raise ValueError("Position less than zero: {}".format(self.position))

        if self.use_direct_query_if_available and isinstance(
                self.notification_log, RecordManagerNotificationLog):
            if advance_by is not None:
                stop_item_num = start_item_num + advance_by
            else:
                stop_item_num = None
            # Directly query for notifications.
            for item in self.notification_log.get_items(start_item_num - 1, stop_item_num):
                yield item
                self.position += 1

        else:
            # Otherwise, use sections (Vaughn Vernon's linked section design).
            section = self.notification_log[self.initial_section_id]

            # Follow previous links.
            while section.previous_id:

                # Break if we can go forward from here.
                if start_item_num is not None:
                    if int(section.section_id.split(',')[0]) <= start_item_num:
                        break

                # Get the previous document.
                section_id = section.previous_id
                section = self.notification_log[section_id]

            # Yield items in first section, optionally after last item number.
            self.section_count += 1
            items = section.items
            if start_item_num is not None:
                section_start_num = int(section.section_id.split(',')[0])
                from_index = start_item_num - section_start_num
                items = items[from_index:]

            if advance_by is None:
                advance_by = -1

            # Yield all items in all subsequent sections.
            while True:

                items_iter = iter(items)
                while True:
                    if stop_index is not None and self.position >= stop_index:
                        return
                    if advance_by == 0:
                        return
                    try:
                        item = next(items_iter)
                    except StopIteration:
                        break
                    self.position += 1
                    advance_by -= 1
                    yield item

                if section.next_id:
                    # Follow link to get next section.
                    section = self.notification_log[section.next_id]
                    items = section.items
                    self.section_count += 1
                else:
                    break

    @property
    def initial_section_id(self):
        """
        Returns initial section ID used to start getting
        linked sections from the notification log.

        Slight departure from Vaughn Vernon's design by not using 'current'
        as initial section ID, but a section ID that just includes the
        "next" position, which the notification log can use to return
        the section containing this position. This avoids lengthy back-
        tracking when reader has a lot of notifications to catch-up on.

        This property has been extracted in order to allow a subclass to
        adjust this default behaviour.

        It would be possible to calculate the actual section ID from the
        current reader position. Using section ID of an actual section
        may hit a cache and avoid troubling the server, but the reader
        would need to know the section size of the notification log it
        is reading. If we don't know section size, perhaps it is a remote
        notification log, we can use 'current' to hit a cache. In future, it
        might be possible to ask the notification log to disclose it's section
        size, or compute an actual section ID for a given position.

        :return: A notification log section ID.
        :rtype: str
        """
        # initial_section_id = 'current'
        if hasattr(self.notification_log, 'section_size'):
            # Todo: Implement this attribute on the remote notification
            #  log class, because that's when we might want to avoid
            #  hitting the server by keeping all section IDs actual
            #  sections that can be cached on the network.
            section_size = self.notification_log.section_size
            # Calculate the section start and end.
            #  - notification ID sequence start,end are 1-based
            #  - 'position' is 1-based, but 1 behind the next position
            #  - we want to get the next position
            #  - the modulo calculation works with 0-based index
            #  - 'position' is equal to zero-based index of next item
            #  - section IDs use 1-based start and end values.
            start = self.position // section_size * section_size
            section_id = '%d,%d' % (start + 1, start + section_size)
        else:
            # Special section ID that indicates next position.
            #  - is used by the notification log to identify the
            #    section which includes the given position
            #  - 'position' is 1-based, but 1 behind the next position
            #  - section IDs use 1-based start and end values.
            section_id = '%d,' % (self.position + 1)

        return section_id

    def read_list(self, advance_by=None):
        return list(self.read_items(advance_by=advance_by))



from eventsourcing.contrib.suffixtrees.domain.model.generalizedsuffixtree import EdgeRepository, \
    GeneralizedSuffixTree, GeneralizedSuffixTreeRepository, NodeRepository, StringidCollection, \
    SuffixTreeEdge, SuffixTreeNode, SuffixTreeNodeChildCollection
from eventsourcing.infrastructure.repositories.collection_repo import CollectionRepository
from eventsourcing.infrastructure.eventsourcedrepository import EventSourcedRepository


class GeneralizedSuffixTreeRepo(EventSourcedRepository, GeneralizedSuffixTreeRepository):
    domain_class = GeneralizedSuffixTree


class NodeRepo(EventSourcedRepository, NodeRepository):
    domain_class = SuffixTreeNode


class NodeChildCollectionRepo(EventSourcedRepository, NodeRepository):
    domain_class = SuffixTreeNodeChildCollection


class EdgeRepo(EventSourcedRepository, EdgeRepository):
    domain_class = SuffixTreeEdge


class StringidCollectionRepo(CollectionRepository):
    domain_class = StringidCollection



from uuid import uuid4

import os
from redis import StrictRedis

from eventsourcing.infrastructure.integersequencegenerators.base import AbstractIntegerSequenceGenerator


class RedisIncr(AbstractIntegerSequenceGenerator):
    """
    Generates a sequence of integers, using Redis' INCR command.

    Maximum number is 2**63, or 9223372036854775807, the maximum
    value of a 64 bit signed integer.
    """
    def __init__(self, redis=None, key=None):
        redis_host = os.getenv('REDIS_HOST', 'localhost')
        self.redis = redis or StrictRedis(host=redis_host)
        self.key = key or 'integer-sequence-generator-{}'.format(uuid4())

    def __next__(self):
        return self.redis.incr(self.key) - 1



from __future__ import absolute_import, division, print_function, unicode_literals

from eventsourcing.domain.model.entity import AbstractEntityRepository, TimestampedVersionedEntity


class Collection(TimestampedVersionedEntity):
    class Event(TimestampedVersionedEntity.Event):
        """Supertype for events of collection entities."""

    class Created(Event, TimestampedVersionedEntity.Created):
        """Published when collection is created."""

    class Discarded(Event, TimestampedVersionedEntity.Discarded):
        """Published when collection is discarded."""

    class EventWithItem(Event):
        @property
        def item(self):
            return self.__dict__['item']

    class ItemAdded(EventWithItem):
        def __mutate__(self, obj):
            obj = super(Collection.ItemAdded, self).__mutate__(obj)
            obj._items.add(self.item)
            return obj

    class ItemRemoved(EventWithItem):
        def __mutate__(self, obj):
            obj = super(Collection.ItemRemoved, self).__mutate__(obj)
            obj._items.remove(self.item)
            return obj

    def __init__(self, **kwargs):
        super(Collection, self).__init__(**kwargs)
        self._items = set()

    def __iter__(self):
        return self._items.__iter__()

    @property
    def items(self):
        self.__assert_not_discarded__()
        return self._items

    def add_item(self, item):
        self.__trigger_event__(self.ItemAdded, item=item)

    def remove_item(self, item):
        self.__trigger_event__(self.ItemRemoved, item=item)


def register_new_collection(collection_id=None):
    return Collection.__create__(originator_id=collection_id)


class AbstractCollectionRepository(AbstractEntityRepository):
    pass



import datetime
from collections import deque
from decimal import Decimal
from unittest import TestCase
from uuid import NAMESPACE_URL

from eventsourcing.utils.times import utc_timezone
from eventsourcing.utils.transcoding import ObjectJSONDecoder, ObjectJSONEncoder, json_loads


class TestObjectJSONEncoder(TestCase):
    def test_encode(self):
        encoder = ObjectJSONEncoder()

        value = 1
        expect = '1'
        self.assertEqual(encoder.encode(value), expect)

        value = datetime.datetime(2011, 1, 1, 1, 1, 1)
        expect = '{"ISO8601_datetime": "2011-01-01T01:01:01.000000"}'
        self.assertEqual(encoder.encode(value), expect)

        value = datetime.datetime(2011, 1, 1, 1, 1, 1, tzinfo=utc_timezone)
        expect = '{"ISO8601_datetime": "2011-01-01T01:01:01.000000+0000"}'
        self.assertEqual(encoder.encode(value), expect)

        value = datetime.date(2011, 1, 1)
        expect = '{"ISO8601_date": "2011-01-01"}'
        self.assertEqual(expect, encoder.encode(value))

        value = datetime.time(23, 59, 59, 123456)
        expect = '{"ISO8601_time": "23:59:59.123456"}'
        self.assertEqual(encoder.encode(value), expect)

        value = Decimal('59.123456')
        expect = '{"__decimal__": "59.123456"}'
        self.assertEqual(encoder.encode(value), expect)

        value = NAMESPACE_URL
        expect = '{"UUID": "6ba7b8119dad11d180b400c04fd430c8"}'
        self.assertEqual(encoder.encode(value), expect)

        value = Object(NAMESPACE_URL)
        self.assertEqual(value.__class__.__module__, 'eventsourcing.tests.test_transcoding',
                         "Module does not have full path, Python 2.7 issue with unittest"
                         " (need to run test from root dir")
        expect = ('{"__class__": {"state": {"a": {"UUID": "6ba7b8119dad11d180b400c04fd430c8"}}, '
                  '"topic": "eventsourcing.tests.test_transcoding#Object"}}')
        self.check_encoded_value(encoder, value, expect)

        value = deque()
        expect = '{"__deque__": []}'
        self.assertEqual(encoder.encode(value), expect)

        # Check defers to base class to raise TypeError.
        # - a type isn't supported at the moment, hence this test works
        with self.assertRaises(TypeError):
            encoder.encode(object)

    def check_encoded_value(self, encoder, value, expect):
        actual = encoder.encode(value)
        try:
            self.assertEqual(expect, actual, (type(expect), type(actual), expect, actual))
        except:
            print()
            print(actual)
            print(expect)


class TestObjectJSONDecoder(TestCase):
    def test_decode(self):
        decoder = ObjectJSONDecoder()
        self.assertEqual(decoder.decode('1'), 1)

        value = '{"ISO8601_datetime": "2011-01-01T01:01:01.000000"}'
        expect = datetime.datetime(2011, 1, 1, 1, 1, 1)
        self.assertEqual(decoder.decode(value), expect)

        value = '{"ISO8601_datetime": "2011-01-01T01:01:01.000000+0000"}'
        expect = datetime.datetime(2011, 1, 1, 1, 1, 1, tzinfo=utc_timezone)
        self.assertEqual(decoder.decode(value), expect)

        value = '{"ISO8601_date": "2011-01-01"}'
        expect = datetime.date(2011, 1, 1)
        self.assertEqual(decoder.decode(value), expect)

        value = '{"UUID": "6ba7b8119dad11d180b400c04fd430c8"}'
        expect = NAMESPACE_URL
        self.assertEqual(decoder.decode(value), expect)

        value = '{"ISO8601_time": "23:59:59.123456"}'
        expect = datetime.time(23, 59, 59, 123456)
        self.assertEqual(decoder.decode(value), expect)

        value = '{"__decimal__": "59.123456"}'
        expect = Decimal('59.123456')
        self.assertEqual(decoder.decode(value), expect)

        value = '{"__deque__": []}'
        expect = deque()
        self.assertEqual(decoder.decode(value), expect)

        value = ('{"__class__": {"state": {"a": {"UUID": "6ba7b8119dad11d180b400c04fd430c8"}}, '
                 '"topic": "eventsourcing.tests.test_transcoding#Object"}}')
        expect = Object(NAMESPACE_URL)
        self.assertEqual(decoder.decode(value), expect)

        # Check raises ValueError when JSON string is invalid.
        with self.assertRaises(ValueError):
            decoder.decode('{')

    def test_json_loads(self):
        # Check raises ValueError when JSON string is invalid.
        with self.assertRaises(ValueError):
            json_loads('{')


class Object(object):
    def __init__(self, a):
        self.a = a

    def __eq__(self, other):
        return self.a == other.a

    def __ne__(self, other):
        return not self.__eq__(other)



from time import sleep, time, process_time
from unittest import skip

from eventsourcing.application.system import SteppingMultiThreadedRunner, System, SteppingSingleThreadedRunner
from eventsourcing.tests.sequenced_item_tests.test_popo_record_manager import PopoTestCase

from eventsourcing.application.popo import PopoApplication
from eventsourcing.tests.test_system import TestSystem
from eventsourcing.tests.test_system_fixtures import create_new_order, Orders, Reservations, Payments


class TestSystemWithPopo(PopoTestCase, TestSystem):
    infrastructure_class = PopoApplication

    def test_singlethreaded_runner_with_multiapp_system(self):
        super(TestSystemWithPopo, self).test_singlethreaded_runner_with_multiapp_system()

    def test_multithreaded_runner_with_singleapp_system(self):
        super(TestSystemWithPopo, self).test_multithreaded_runner_with_singleapp_system()

    def test_multithreaded_runner_with_multiapp_system(self):
        super(TestSystemWithPopo, self).test_multithreaded_runner_with_multiapp_system()

    def test_clocked_multithreaded_runner_with_multiapp_system(self):
        super(TestSystemWithPopo, self).test_clocked_multithreaded_runner_with_multiapp_system()

    def test_stepping_singlethreaded_runner_with_multiapp_system(self):
        system = System(
            Orders | Reservations | Orders,
            Orders | Payments | Orders,
            setup_tables=True,
            infrastructure_class=self.infrastructure_class
        )

        self.set_db_uri()

        normal_speed = 3
        scale_factor = 0
        runner = SteppingSingleThreadedRunner(
            system=system,
            normal_speed=normal_speed,
            scale_factor=scale_factor,
            # is_verbose=True,
        )

        start_time = time()
        start_process_time = process_time()

        with runner:

            orders = system.processes['orders']

            # Create a new order.
            num_orders = 120
            order_ids = []
            for i in range(num_orders):
                def cmd():
                    order_id = create_new_order()
                    assert order_id in orders.repository
                    order_ids.append(order_id)
                runner.call_in_future(cmd, 20 * (i + 1))
                # sleep(0.1)
                # sleep(tick_interval / 3)
                # sleep(tick_interval * 10)

            retries = 10 * num_orders
            while len(order_ids) < num_orders:
                sleep(0.01)

            for order_id in order_ids[-1:]:
                # while not orders.repository[order_id].is_reserved:
                #     sleep(0.1)
                #     retries -= 1
                #     assert retries, "Failed set order.is_reserved"

                while retries and not orders.repository[order_id].is_paid:
                    if not runner.clock_thread.is_alive():
                        self.fail("Clock thread died")
                    # else:
                    #     print("clock thread is alive")

                    # sleep(1)
                    sleep(0.2)
                    retries -= 1

        if runner.clock_thread.is_alive():
            self.fail("Clock thread still alive")
        final_time = runner.clock_thread.tick_count / runner.normal_speed
        print(f"Runner: average clock speed {runner.clock_thread.actual_clock_speed:.1f}Hz")
        print(f"Runner: total tick count {runner.clock_thread.tick_count}")
        print(f"Runner: total time in simulation {final_time:.2f}s")

        elapsed_time = time() - start_time
        print(f"Duration: { elapsed_time :.4f}s")
        execution_time = process_time() - start_process_time
        print(f"CPU: { 100 * execution_time / elapsed_time :.2f}%")

        assert retries, "Failed set order.is_paid"
        sleep(0.001)

    def test_stepping_multithreaded_runner_with_multiapp_system(self):
        system = System(
            Orders | Reservations | Orders,
            Orders | Payments | Orders,
            setup_tables=True,
            infrastructure_class=self.infrastructure_class
        )

        self.set_db_uri()

        normal_speed = 5
        scale_factor = 0
        runner = SteppingMultiThreadedRunner(
            system=system,
            normal_speed=normal_speed,
            scale_factor=scale_factor,
            is_verbose=False,
        )
        with runner:

            start_time = time()
            start_process_time = process_time()

            orders = system.processes['orders']

            # Create a new order.
            num_orders = 12
            order_ids = []
            for i in range(num_orders):
                def cmd():
                    order_id = create_new_order()
                    assert order_id in orders.repository
                    order_ids.append(order_id)
                runner.call_in_future(cmd, 20 * (i + 1))
                # sleep(0.1)
                # sleep(tick_interval / 3)
                # sleep(tick_interval * 10)

            while len(order_ids) < num_orders:
                sleep(0.01)

            retries = 10 * num_orders
            for order_id in order_ids[-1:]:
                # while not orders.repository[order_id].is_reserved:
                #     sleep(0.1)
                #     retries -= 1
                #     assert retries, "Failed set order.is_reserved"

                while retries and not orders.repository[order_id].is_paid:
                    # sleep(1)
                    sleep(0.2)
                    retries -= 1

        final_time = runner.clock_thread.tick_count / runner.normal_speed
        print(f"Runner: average clock speed {runner.clock_thread.actual_clock_speed:.0f}Hz")
        print(f"Runner: total tick count {runner.clock_thread.tick_count}")
        print(f"Runner: total time in simulation {final_time:.2f}s")

        elapsed_time = time() - start_time
        print(f"Duration: { elapsed_time :.4f}s")
        execution_time = process_time() - start_process_time
        print(f"CPU: { 100 * execution_time / elapsed_time :.2f}%")

        assert retries, "Failed set order.is_paid"

    @skip("Popo record manager doesn't support multiprocessing")
    def test_multiprocessing_multiapp_system(self):
        super(TestSystemWithPopo, self).test_multiprocessing_multiapp_system()

    @skip("Popo record manager doesn't support multiprocessing")
    def test_multiprocessing_singleapp_system(self):
        super(TestSystemWithPopo, self).test_multiprocessing_singleapp_system()

    @skip("Popo record manager doesn't support multiprocessing")
    def test_multipipeline_multiprocessing_multiapp(self):
        super(TestSystemWithPopo, self).test_multipipeline_multiprocessing_multiapp()

    def set_db_uri(self):
        # The Popo settings module doesn't currently recognise DB_URI.
        pass


# Avoid running imported test case.
del TestSystem



import logging

from thespian.actors import *

from eventsourcing.application.notificationlog import RecordManagerNotificationLog
from eventsourcing.application.process import ProcessApplication, Prompt
from eventsourcing.application.system import System, SystemRunner
from eventsourcing.domain.model.events import subscribe, unsubscribe
from eventsourcing.exceptions import RecordConflictError

logger = logging.getLogger()

# Todo: Send timer message to run slave every so often (in master or slave?).


DEFAULT_ACTORS_LOGCFG = {
    'version': 1,
    'formatters': {
        'normal': {
            'format': '%(levelname)-8s %(message)s'
        }
    },
    'handlers': {
        # 'h': {
        #     'class': 'logging.FileHandler',
        #     'filename': 'hello.log',
        #     'formatter': 'normal',
        #     'level': logging.INFO
        # }
    },
    'loggers': {
        # '': {'handlers': ['h'], 'level': logging.DEBUG}
    }
}


def start_actor_system(system_base=None, logcfg=DEFAULT_ACTORS_LOGCFG):
    ActorSystem(
        systemBase=system_base,
        logDefs=logcfg,
    )


def shutdown_actor_system():
    ActorSystem().shutdown()


def start_multiproc_tcp_base_system():
    start_actor_system(system_base='multiprocTCPBase')


# def start_multiproc_udp_base_system():
#     start_actor_system(system_base='multiprocUDPBase')
#
#
# def start_multiproc_queue_base_system():
#     start_actor_system(system_base='multiprocQueueBase')


class ActorModelRunner(SystemRunner):
    """
    Uses actor model framework to run a system of process applications.
    """
    def __init__(self, system: System, pipeline_ids, system_actor_name='system', shutdown_on_close=False, **kwargs):
        super(ActorModelRunner, self).__init__(system=system, **kwargs)
        self.pipeline_ids = list(pipeline_ids)
        self.pipeline_actors = {}
        self.system_actor_name = system_actor_name
        # Create the system actor (singleton).
        self.system_actor = self.actor_system.createActor(
            actorClass=SystemActor,
            globalName=self.system_actor_name
        )
        self.shutdown_on_close = shutdown_on_close

    @property
    def actor_system(self):
        return ActorSystem()

    def start(self):
        """
        Starts all the actors to run a system of process applications.
        """
        # Subscribe to broadcast prompts published by a process
        # application in the parent operating system process.
        subscribe(handler=self.forward_prompt, predicate=self.is_prompt)

        # Initialise the system actor.
        msg = SystemInitRequest(
            self.system.process_classes,
            self.infrastructure_class,
            self.system.followings,
            self.pipeline_ids
        )
        response = self.actor_system.ask(self.system_actor, msg)

        # Keep the pipeline actor addresses, to send prompts directly.
        assert isinstance(response, SystemInitResponse), type(response)

        assert list(response.pipeline_actors.keys()) == self.pipeline_ids, (
            "Configured pipeline IDs mismatch initialised system {} {}").format(
            list(self.pipeline_actors.keys()), self.pipeline_ids
        )

        self.pipeline_actors = response.pipeline_actors

        # Todo: Somehow know when to get a new address from the system actor.
        # Todo: Command and response messages to system actor to get new pipeline address.

    @staticmethod
    def is_prompt(event):
        return isinstance(event, Prompt)

    def forward_prompt(self, prompt):
        if prompt.pipeline_id in self.pipeline_actors:
            pipeline_actor = self.pipeline_actors[prompt.pipeline_id]
            self.actor_system.tell(pipeline_actor, prompt)
        # else:
        #     msg = "Pipeline {} is not running.".format(prompt.pipeline_id)
        #     raise ValueError(msg)

    def close(self):
        """Stops all the actors running a system of process applications."""
        super(ActorModelRunner, self).close()
        unsubscribe(handler=self.forward_prompt, predicate=self.is_prompt)
        if self.shutdown_on_close:
            self.shutdown()

    def shutdown(self):
        msg = ActorExitRequest(recursive=True)
        self.actor_system.tell(self.system_actor, msg)


class SystemActor(Actor):
    def __init__(self):
        super(SystemActor, self).__init__()
        self.pipeline_actors = {}
        self.is_initialised = False

    def receiveMessage(self, msg, sender):
        if isinstance(msg, SystemInitRequest):
            if not self.is_initialised:
                self.init_pipelines(msg)
                self.is_initialised = True
            msg = SystemInitResponse(self.pipeline_actors.copy())
            self.send(sender, msg)

    def init_pipelines(self, msg):
        self.process_classes = msg.process_classes
        self.infrastructure_class = msg.infrastructure_class
        self.system_followings = msg.system_followings
        for pipeline_id in msg.pipeline_ids:
            pipeline_actor = self.createActor(PipelineActor)
            self.pipeline_actors[pipeline_id] = pipeline_actor
            msg = PipelineInitRequest(
                self.process_classes,
                self.infrastructure_class,
                self.system_followings,
                pipeline_id
            )
            self.send(pipeline_actor, msg)


class PipelineActor(Actor):
    def __init__(self):
        super(PipelineActor, self).__init__()
        self.system = None
        self.process_actors = {}
        self.pipeline_id = None

    def receiveMessage(self, msg, sender):
        if isinstance(msg, PipelineInitRequest):
            # logger.info("pipeline received init: {}".format(msg))
            self.init_pipeline(msg)
        elif isinstance(msg, Prompt):
            # logger.info("pipeline received prompt: {}".format(msg))
            self.forward_prompt(msg)

    def init_pipeline(self, msg):
        self.pipeline_id = msg.pipeline_id
        self.process_classes = msg.process_classes
        self.infrastructure_class = msg.infrastructure_class
        self.system_followings = msg.system_followings

        self.followers = {}
        for process_class_name, upstream_class_names in self.system_followings.items():
            for upstream_class_name in upstream_class_names:
                process_name = upstream_class_name.lower()
                if process_name not in self.followers:
                    self.followers[process_name] = []
                downstream_class_names = self.followers[process_name]
                if process_class_name not in downstream_class_names:
                    downstream_class_names.append(process_class_name)

        process_class_names = self.system_followings.keys()
        for process_class_name in process_class_names:
            process_actor = self.createActor(ProcessMaster)
            process_name = process_class_name.lower()
            self.process_actors[process_name] = process_actor

        for process_class_name in process_class_names:
            process_name = process_class_name.lower()
            upstream_application_names = [c.lower() for c in self.system_followings[process_class_name]]
            downstream_actors = {}
            for downstream_class_name in self.followers[process_name]:
                downstream_name = downstream_class_name.lower()
                # logger.warning("sending prompt to process application {}".format(downstream_name))
                process_actor = self.process_actors[downstream_name]
                downstream_actors[downstream_name] = process_actor
            process_class = self.process_classes[process_class_name]
            msg = ProcessInitRequest(
                process_class,
                self.infrastructure_class,
                self.pipeline_id,
                upstream_application_names,
                downstream_actors,
                self.myAddress
            )
            self.send(self.process_actors[process_name], msg)

    def forward_prompt(self, msg):
        for downstream_class_name in self.followers[msg.process_name]:
            downstream_name = downstream_class_name.lower()
            process_actor = self.process_actors[downstream_name]
            self.send(process_actor, msg)


class ProcessMaster(Actor):
    def __init__(self):
        super(ProcessMaster, self).__init__()
        self.is_slave_running = False
        self.last_prompts = {}
        self.slave_actor = None

    def receiveMessage(self, msg, sender):
        if isinstance(msg, ProcessInitRequest):
            self.init_process(msg)
        elif isinstance(msg, Prompt):
            # logger.warning("{} master received prompt: {}".format(self.process_application_class.__name__, msg))
            self.consume_prompt(prompt=msg)
        elif isinstance(msg, SlaveRunResponse):
            # logger.info("process application master received slave finished run: {}".format(msg))
            self.handle_slave_run_response()

    def init_process(self, msg):
        self.process_application_class = msg.process_application_class
        self.infrastructure_class = msg.infrastructure_class
        self.slave_actor = self.createActor(ProcessSlave)
        self.send(self.slave_actor, msg)
        self.run_slave()

    def consume_prompt(self, prompt):
        self.last_prompts[prompt.process_name] = prompt
        self.run_slave()

    def handle_slave_run_response(self):
        self.is_slave_running = False
        if self.last_prompts:
            self.run_slave()

    def run_slave(self):
        # Don't send to slave if we think it's running, or we'll
        # probably get blocked while sending the message and have
        # to wait until the slave runs its loop (thespian design).
        if self.slave_actor and not self.is_slave_running:
            self.send(self.slave_actor, SlaveRunRequest(self.last_prompts, self.myAddress))
            self.is_slave_running = True
            self.last_prompts = {}


class ProcessSlave(Actor):
    def __init__(self):
        super(ProcessSlave, self).__init__()
        self.process = None

    def receiveMessage(self, msg, sender):
        if isinstance(msg, ProcessInitRequest):
            # logger.info("process application slave received init: {}".format(msg))
            self.init_process(msg)
        elif isinstance(msg, SlaveRunRequest):
            # logger.info("{} process application slave received last prompts: {}".format(self.process.name, msg))
            self.run_process(msg)
        elif isinstance(msg, ActorExitRequest):
            # logger.info("{} process application slave received exit request: {}".format(self.process.name, msg))
            self.close()

    def init_process(self, msg):
        self.pipeline_actor = msg.pipeline_actor
        self.downstream_actors = msg.downstream_actors
        self.pipeline_id = msg.pipeline_id
        self.upstream_application_names = msg.upstream_application_names

        # Construct the process application class.
        process_class = msg.process_application_class
        if msg.infrastructure_class:
            process_class = process_class.mixin(msg.infrastructure_class)

        # Reset the database connection (for Django).
        process_class.reset_connection_after_forking()

        # Construct the process application.
        self.process = process_class(
            pipeline_id=self.pipeline_id,
        )
        assert isinstance(self.process, ProcessApplication)

        # Subscribe the slave actor's send_prompt() method.
        #  - the process application will call publish_prompt()
        #    and the actor will receive the prompt and send it
        #    as a message.
        subscribe(
            predicate=self.is_my_prompt,
            handler=self.send_prompt
        )

        # Close the process application persistence policy.
        #  - slave actor process application doesn't publish
        #    events, so we don't need this
        self.process.persistence_policy.close()

        # Unsubscribe process application's publish_prompt().
        #  - slave actor process application doesn't publish
        #    events, so we don't need this
        unsubscribe(
            predicate=self.process.persistence_policy.is_event,
            handler=self.process.publish_prompt
        )


        # Construct and follow upstream notification logs.
        for upstream_application_name in self.upstream_application_names:
            record_manager = self.process.event_store.record_manager
            # assert isinstance(record_manager, ACIDRecordManager), type(record_manager)
            notification_log = RecordManagerNotificationLog(
                record_manager=record_manager.clone(
                    application_name=upstream_application_name,
                    pipeline_id=self.pipeline_id
                ),
                section_size=self.process.notification_log_section_size
            )
            self.process.follow(upstream_application_name, notification_log)

    def run_process(self, msg):
        notification_count = 0
        # Just process one notification so prompts are dispatched promptly, sent
        # messages only dispatched from actor after receive_message() returns.
        advance_by = 1

        try:
            if msg.last_prompts:
                for prompt in msg.last_prompts.values():
                    notification_count += self.process.run(prompt, advance_by=advance_by)
            else:
                notification_count += self.process.run(advance_by=advance_by)
        except RecordConflictError:
            # Run again.
            self.send(self.myAddress, SlaveRunRequest(last_prompts={}, master=msg.master))
        else:
            if notification_count:
                # Run again, until nothing was done.
                self.send(self.myAddress, SlaveRunRequest(last_prompts={}, master=msg.master))
            else:
                # Report back to master.
                self.send(msg.master, SlaveRunResponse())

    def close(self):
        unsubscribe(
            predicate=self.is_my_prompt,
            handler=self.send_prompt
        )

        self.process.close()

    def is_my_prompt(self, prompt):
        return (
            isinstance(prompt, Prompt)
            and prompt.process_name == self.process.name
            and prompt.pipeline_id == self.pipeline_id
        )

    def send_prompt(self, prompt):
        for downstream_name, downstream_actor in self.downstream_actors.items():
            self.send(downstream_actor, prompt)


class SystemInitRequest(object):
    def __init__(self, process_classes, infrastructure_class, system_followings, pipeline_ids):
        self.process_classes = process_classes
        self.infrastructure_class = infrastructure_class
        self.system_followings = system_followings
        self.pipeline_ids = pipeline_ids


class SystemInitResponse(object):
    def __init__(self, pipeline_actors):
        self.pipeline_actors = pipeline_actors


class PipelineInitRequest(object):
    def __init__(self, process_classes, infrastructure_class, system_followings, pipeline_id):
        self.process_classes = process_classes
        self.infrastructure_class = infrastructure_class
        self.system_followings = system_followings
        self.pipeline_id = pipeline_id


class ProcessInitRequest(object):
    def __init__(self, process_application_class, infrastructure_class, pipeline_id,
                 upstream_application_names,
                 downstream_actors,
                 pipeline_actor):
        self.process_application_class = process_application_class
        self.infrastructure_class = infrastructure_class
        self.pipeline_id = pipeline_id
        self.upstream_application_names = upstream_application_names
        self.downstream_actors = downstream_actors
        self.pipeline_actor = pipeline_actor


class SlaveRunRequest(object):
    def __init__(self, last_prompts, master):
        self.last_prompts = last_prompts
        self.master = master


class SlaveRunResponse(object):
    pass



from sqlalchemy import DECIMAL, String
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.sql.schema import Column, Index
from sqlalchemy.sql.sqltypes import BigInteger, Integer, Text
from sqlalchemy_utils.types.uuid import UUIDType

Base = declarative_base()


# Please note, the record classes without an indexed ID ('WithID') are
# more or less equivalent to the similarly named Django classes. The
# record classes without ID ('NoID') are more or less equivalent to the
# Cassandra classes, in that they don't have a record ID, so must be
# accessed by sequence ID, optionally with position, and also their
# application sequence must be constructed elsewhere such as with a big
# array. Without record IDs, the maximum rate at which events can be
# written in parallel will be greater, because the IDs don't need to be
# generated and because there isn't a record ID index to be updated.
# Also, if there is no record ID index, the table can easily be shared
# with the sequence ID being used to select a shard. The drawback is
# that the application sequence will need to be constructed elsewhere.


class IntegerSequencedWithIDRecord(Base):
    __tablename__ = 'integer_sequenced_items'

    # Record ID.
    id = Column(BigInteger().with_variant(Integer, "sqlite"), primary_key=True, index=True, unique=True)

    # Sequence ID (e.g. an entity or aggregate ID).
    sequence_id = Column(UUIDType(), nullable=False)

    # Position (index) of item in sequence.
    position = Column(BigInteger().with_variant(Integer, "sqlite"), nullable=False)

    # Topic of the item (e.g. path to domain event class).
    topic = Column(Text(), nullable=False)

    # State of the item (serialized dict, possibly encrypted).
    state = Column(Text())

    __table_args__ = (
        Index('integer_sequenced_items_sequence_id_position_index', 'sequence_id', 'position', unique=True),
    )


class IntegerSequencedNoIDRecord(Base):
    __tablename__ = 'integer_sequenced_items_noid'

    # Sequence ID (e.g. an entity or aggregate ID).
    sequence_id = Column(UUIDType(), primary_key=True)

    # Position (index) of item in sequence.
    position = Column(BigInteger().with_variant(Integer, "sqlite"), primary_key=True)

    # Topic of the item (e.g. path to domain event class).
    topic = Column(Text(), nullable=False)

    # State of the item (serialized dict, possibly encrypted).
    state = Column(Text())


IntegerSequencedRecord = IntegerSequencedWithIDRecord


class TimestampSequencedWithIDRecord(Base):
    __tablename__ = 'timestamp_sequenced_items'

    # Record ID.
    id = Column(BigInteger().with_variant(Integer, "sqlite"), primary_key=True, index=True, unique=True,
                autoincrement=True)

    # Sequence ID (e.g. an entity or aggregate ID).
    sequence_id = Column(UUIDType(), nullable=False)

    # Position (timestamp) of item in sequence.
    position = Column(DECIMAL(24, 6, 6), nullable=False, unique=False)

    # Topic of the item (e.g. path to domain event class).
    topic = Column(Text(), nullable=False)

    # State of the item (serialized dict, possibly encrypted).
    state = Column(Text())

    __table_args__ = (
        Index('timestamp_sequenced_items_sequence_id_position_index', 'sequence_id', 'position', unique=True),
        Index('timestamp_sequenced_items_position_index', 'position', unique=False),
    )


class TimestampSequencedNoIDRecord(Base):
    __tablename__ = 'timestamp_sequenced_items_noid'

    # Sequence ID (e.g. an entity or aggregate ID).
    sequence_id = Column(UUIDType(), primary_key=True)

    # Position (timestamp) of item in sequence.
    position = Column(DECIMAL(24, 6, 6), primary_key=True)

    # Topic of the item (e.g. path to domain event class).
    topic = Column(Text(), nullable=False)

    # State of the item (serialized dict, possibly encrypted).
    state = Column(Text())

    __table_args__ = (
        Index('timestamp_sequenced_items_noid_position_index', 'position', unique=False),
    )


TimestampSequencedRecord = TimestampSequencedNoIDRecord


class SnapshotRecord(Base):
    __tablename__ = 'snapshots'

    # Sequence ID (e.g. an entity or aggregate ID).
    sequence_id = Column(UUIDType(), primary_key=True)

    # Position (index) of item in sequence.
    position = Column(BigInteger().with_variant(Integer, "sqlite"), primary_key=True)

    # Topic of the item (e.g. path to domain entity class).
    topic = Column(Text(), nullable=False)

    # State of the item (serialized dict, possibly encrypted).
    state = Column(Text())


class EntitySnapshotRecord(Base):
    __tablename__ = 'entity_snapshots'

    # Application ID.
    application_name = Column(String(length=32), primary_key=True)

    # Originator ID (e.g. an entity or aggregate ID).
    originator_id = Column(UUIDType(), primary_key=True)

    # Originator version of item in sequence.
    originator_version = Column(BigInteger().with_variant(Integer, "sqlite"), primary_key=True)

    # Topic of the item (e.g. path to domain entity class).
    topic = Column(Text(), nullable=False)

    # State of the item (serialized dict, possibly encrypted).
    state = Column(Text())


class StoredEventRecord(Base):
    __tablename__ = 'stored_events'

    # Application ID.
    application_name = Column(String(length=32), primary_key=True)

    # Originator ID (e.g. an entity or aggregate ID).
    originator_id = Column(UUIDType(), primary_key=True)

    # Originator version of item in sequence.
    originator_version = Column(BigInteger().with_variant(Integer, "sqlite"), primary_key=True)

    # Pipeline ID.
    pipeline_id = Column(Integer(), nullable=True)

    # Notification ID.
    notification_id = Column(BigInteger().with_variant(Integer, "sqlite"), nullable=True)

    # Topic of the item (e.g. path to domain event class).
    topic = Column(Text(), nullable=False)

    # State of the item (serialized dict, possibly encrypted).
    state = Column(Text())

    # Causal dependencies.
    causal_dependencies = Column(Text())

    __table_args__ = (
        Index(
            'stored_events_notification_index',
            'application_name',
            'pipeline_id',
            'notification_id',
            unique=True,
        ),
    )


class NotificationTrackingRecord(Base):
    __tablename__ = 'notification_tracking'

    # Application name.
    application_name = Column(String(length=32), primary_key=True)

    # Upstream application name.
    upstream_application_name = Column(String(length=32), primary_key=True)

    # Pipeline ID.
    pipeline_id = Column(Integer(), primary_key=True)

    # Notification ID.
    notification_id = Column(BigInteger().with_variant(Integer, "sqlite"), primary_key=True)



import multiprocessing
from multiprocessing import Manager
from queue import Empty
from time import sleep

from eventsourcing.application.notificationlog import RecordManagerNotificationLog
from eventsourcing.application.process import Prompt
from eventsourcing.application.simple import ApplicationWithConcreteInfrastructure
from eventsourcing.application.system import DEFAULT_POLL_INTERVAL, PromptOutbox, System, SystemRunner
from eventsourcing.domain.model.decorators import retry
from eventsourcing.domain.model.events import subscribe, unsubscribe
from eventsourcing.exceptions import CausalDependencyFailed, OperationalError, RecordConflictError, ProgrammingError
from eventsourcing.infrastructure.base import DEFAULT_PIPELINE_ID


class MultiprocessRunner(SystemRunner):

    def __init__(self, system: System, pipeline_ids=(DEFAULT_PIPELINE_ID,), poll_interval=None,
                 setup_tables=False, sleep_for_setup_tables=0, *args, **kwargs):
        super(MultiprocessRunner, self).__init__(system=system, *args, **kwargs)
        self.pipeline_ids = pipeline_ids
        self.poll_interval = poll_interval or DEFAULT_POLL_INTERVAL
        assert isinstance(system, System)
        self.os_processes = None
        self.setup_tables = setup_tables or system.setup_tables
        self.sleep_for_setup_tables = sleep_for_setup_tables

    def start(self):
        assert self.os_processes is None, "Already started"

        self.os_processes = []

        self.manager = Manager()
        self.inboxes = {}
        self.outboxes = {}

        # Setup queues.
        for pipeline_id in self.pipeline_ids:
            for process_name, upstream_names in self.system.followings.items():
                inbox_id = (pipeline_id, process_name.lower())
                if inbox_id not in self.inboxes:
                    self.inboxes[inbox_id] = self.manager.Queue()
                for upstream_class_name in upstream_names:
                    outbox_id = (pipeline_id, upstream_class_name.lower())
                    if outbox_id not in self.outboxes:
                        self.outboxes[outbox_id] = PromptOutbox()
                    if inbox_id not in self.outboxes[outbox_id].downstream_inboxes:
                        self.outboxes[outbox_id].downstream_inboxes[inbox_id] = self.inboxes[inbox_id]

        # Subscribe to broadcast prompts published by a process
        # application in the parent operating system process.
        subscribe(handler=self.broadcast_prompt, predicate=self.is_prompt)

        # Start operating system process.
        for pipeline_id in self.pipeline_ids:
            for process_name, upstream_names in self.system.followings.items():
                process_class = self.system.process_classes[process_name]
                inbox = self.inboxes[(pipeline_id, process_name.lower())]
                outbox = self.outboxes.get((pipeline_id, process_name.lower()))
                os_process = OperatingSystemProcess(
                    application_process_class=process_class,
                    infrastructure_class=self.infrastructure_class,
                    upstream_names=upstream_names,
                    poll_interval=self.poll_interval,
                    pipeline_id=pipeline_id,
                    setup_tables=self.setup_tables,
                    inbox=inbox,
                    outbox=outbox,
                )
                os_process.daemon = True
                os_process.start()
                self.os_processes.append(os_process)
                if self.setup_tables:
                    # Avoid conflicts when creating tables.
                    sleep(self.sleep_for_setup_tables)

    def broadcast_prompt(self, prompt):
        outbox_id = (prompt.pipeline_id, prompt.process_name)
        outbox = self.outboxes.get(outbox_id)
        if outbox:
            outbox.put(prompt)

    @staticmethod
    def is_prompt(event):
        return isinstance(event, Prompt)

    def close(self):
        super(MultiprocessRunner, self).close()

        unsubscribe(handler=self.broadcast_prompt, predicate=self.is_prompt)

        for os_process in self.os_processes:
            os_process.inbox.put('QUIT')

        for os_process in self.os_processes:
            os_process.join(timeout=10)

        for os_process in self.os_processes:
            os_process.is_alive() and os_process.terminate()

        self.os_processes = None
        self.manager = None


class OperatingSystemProcess(multiprocessing.Process):

    def __init__(self, application_process_class, infrastructure_class, upstream_names,
                 pipeline_id=DEFAULT_PIPELINE_ID, poll_interval=DEFAULT_POLL_INTERVAL,
                 setup_tables=False, inbox=None, outbox=None, *args, **kwargs):
        super(OperatingSystemProcess, self).__init__(*args, **kwargs)
        self.application_process_class = application_process_class
        self.infrastructure_class = infrastructure_class
        self.upstream_names = upstream_names
        self.daemon = True
        self.pipeline_id = pipeline_id
        self.poll_interval = poll_interval
        self.inbox = inbox
        self.outbox = outbox
        self.setup_tables = setup_tables

    def run(self):
        # Construct process application class.
        process_class = self.application_process_class
        if not isinstance(process_class, ApplicationWithConcreteInfrastructure):
            if self.infrastructure_class:
                process_class = process_class.mixin(self.infrastructure_class)
            else:
                raise ProgrammingError('infrastructure_class is not set')

        # Construct process application object.
        self.process = process_class(
            pipeline_id=self.pipeline_id,
            setup_table=self.setup_tables,
        )

        # Follow upstream notification logs.
        for upstream_name in self.upstream_names:

            # Obtain a notification log object (local or remote) for the upstream process.
            if upstream_name == self.process.name:
                # Upstream is this process's application,
                # so use own notification log.
                notification_log = self.process.notification_log
            else:
                # For a different application, we need to construct a notification
                # log with a record manager that has the upstream application ID.
                # Currently assumes all applications are using the same database
                # and record manager class. If it wasn't the same database,we would
                # to use a remote notification log, and upstream would need to provide
                # an API from which we can pull. It's not unreasonable to have a fixed
                # number of application processes connecting to the same database.
                record_manager = self.process.event_store.record_manager

                notification_log = RecordManagerNotificationLog(
                    record_manager=record_manager.clone(
                        application_name=upstream_name,
                        # Todo: Check if setting pipeline_id is necessary (it's the same?).
                        pipeline_id=self.pipeline_id
                    ),
                    section_size=self.process.notification_log_section_size
                )
                # Todo: Support upstream partition IDs different from self.pipeline_id?
                # Todo: Support combining partitions. Read from different partitions but write to the same partition,
                # could be one os process that reads from many logs of the same upstream app, or many processes each
                # reading one partition with contention writing to the same partition).
                # Todo: Support dividing partitions Read from one but write to many. Maybe one process per
                # upstream partition, round-robin to pick partition for write. Or have many processes reading
                # with each taking it in turn to skip processing somehow.
                # Todo: Dividing partitions would allow a stream to flow at the same rate through slower
                # process applications.
                # Todo: Support merging results from "replicated state machines" - could have a command
                # logging process that takes client commands and presents them in a notification log.
                # Then the system could be deployed in different places, running independently, receiving
                # the same commands, and running the same processes. The command logging process could
                # be accompanied with a result logging process that reads results from replicas as they
                # are available. Not sure what to do if replicas return different things. If one replica
                # goes down, then it could resume by pulling events from another? Not sure what to do.
                # External systems could be modelled as commands.

            # Make the process follow the upstream notification log.
            self.process.follow(upstream_name, notification_log)

        # Subscribe to broadcast prompts published by the process application.
        subscribe(handler=self.broadcast_prompt, predicate=self.is_prompt)

        try:
            self.loop_on_prompts()
        finally:
            unsubscribe(handler=self.broadcast_prompt, predicate=self.is_prompt)

    @retry(CausalDependencyFailed, max_attempts=100, wait=0.1)
    def loop_on_prompts(self):

        # Run once, in case prompts were missed.
        self.run_process()

        # Loop on getting prompts.
        while True:
            try:
                # Todo: Make the poll interval gradually increase if there are only timeouts?
                item = self.inbox.get(timeout=self.poll_interval)
                self.inbox.task_done()

                if item == 'QUIT':
                    self.process.close()
                    break

                else:
                    self.run_process(item)

            except Empty:
                # Basically, we're polling after a timeout.
                self.run_process()

    @retry((OperationalError, RecordConflictError), max_attempts=100, wait=0.1)
    def run_process(self, prompt=None):
        self.process.run(prompt)

    def broadcast_prompt(self, prompt):
        if self.outbox is not None:
            self.outbox.put(prompt)

    @staticmethod
    def is_prompt(event):
        return isinstance(event, Prompt)



from eventsourcing.tests.example_application_tests import base
from eventsourcing.tests.sequenced_item_tests.test_sqlalchemy_record_manager import \
    SQLAlchemyRecordManagerTestCase


class TestExampleApplicationWithSQLAlchemy(SQLAlchemyRecordManagerTestCase, base.ExampleApplicationTestCase):
    pass
